from mage_ai.data_preparation.executors.block_executor import BlockExecutor
from mage_ai.services.k8s.job_manager import JobManager as K8sJobManager
from mage_ai.shared.hash import merge_dict
from typing import Dict


class K8sBlockExecutor(BlockExecutor):
    def __init__(self, pipeline, block_uuid: str, execution_partition: str = None):
        super().__init__(pipeline, block_uuid, execution_partition=execution_partition)
        self.executor_config = self.pipeline.repo_config.k8s_executor_config or dict()
        if self.block.executor_config is not None:
            self.executor_config = merge_dict(self.executor_config, self.block.executor_config)

    def _execute(
        self,
        block_run_id: int = None,
        global_vars: Dict = None,
        **kwargs,
    ) -> None:
        job_manager = K8sJobManager(
            job_name=f'mage-data-prep-block-{block_run_id}',
            logger=self.logger,
            logging_tags=kwargs.get('tags', dict()),
        )
        cmd = self._run_commands(block_run_id, global_vars, **kwargs)
        job_manager.run_job(
            cmd,
            k8s_config=self.executor_config,
        )



from mage_ai.data_preparation.git import Git
from mage_ai.data_preparation.sync import GitConfig
from mage_ai.data_preparation.sync.base_sync import BaseSync
from mage_ai.shared.logger import VerboseFunctionExec


class GitSync(BaseSync):
    def __init__(self, sync_config: GitConfig):
        self.branch = sync_config.branch
        self.remote_repo_link = sync_config.remote_repo_link
        self.git_manager = Git(sync_config)

    def sync_data(self):
        with VerboseFunctionExec(
            f'Syncing data with remote repo {self.remote_repo_link}',
            verbose=True,
        ):
            self.git_manager.reset(self.branch)



from datetime import datetime
from mage_ai.orchestration.db import safe_db_query
from mage_ai.orchestration.db.errors import ValidationError
from mage_ai.orchestration.db.models.base import BaseModel
from sqlalchemy import (
    Column,
    Boolean,
    DateTime,
    Enum,
    ForeignKey,
    Integer,
    JSON,
    String,
)
from sqlalchemy.orm import relationship, validates
import enum
import re


class User(BaseModel):
    avatar = Column(String(255), default=None)
    email = Column(String(255), default=None, index=True, unique=True)
    first_name = Column(String(255), default=None)
    last_name = Column(String(255), default=None)
    owner = Column(Boolean, default=False)
    password_hash = Column(String(255), default=None)
    password_salt = Column(String(255), default=None)
    roles = Column(Integer, default=None)
    username = Column(String(255), default=None, index=True, unique=True)
    preferences = Column(JSON, default=None)

    oauth2_applications = relationship('Oauth2Application', back_populates='user')
    oauth2_access_tokens = relationship('Oauth2AccessToken', back_populates='user')

    @validates('email')
    def validate_email(self, key, value):
        if value:
            regex = re.compile(r"([-!#-'*+/-9=?A-Z^-~]+(\.[-!#-'*+/-9=?A-Z^-~]+)*|\"([]!#-[^-~ \t]|(\\[\t -~]))+\")@([-!#-'*+/-9=?A-Z^-~]+(\.[-!#-'*+/-9=?A-Z^-~]+)*|\[[\t -Z^-~]*])")  # noqa: E501
            if not re.fullmatch(regex, value):
                raise ValidationError('Email address is invalid.', metadata=dict(
                    key=key,
                    value=value,
                ))
        return value

    @property
    def roles_display(self) -> str:
        if self.owner:
            return 'Owner'
        elif self.roles:
            if self.roles & 1 != 0:
                return 'Admin'
            elif self.roles & 2 != 0:
                return 'Editor'
            elif self.roles & 4 != 0:
                return 'Viewer'


class Oauth2Application(BaseModel):
    class AuthorizationGrantType(str, enum.Enum):
        AUTHORIZATION_CODE = 'authorization-code'
        CLIENT_CREDENTIALS = 'client-credentials'

    class ClientType(str, enum.Enum):
        PRIVATE = 'private'
        PUBLIC = 'public'

    authorization_grant_type = Column(
        Enum(AuthorizationGrantType),
        default=AuthorizationGrantType.AUTHORIZATION_CODE,
    )
    client_id = Column(String(255), index=True, unique=True)
    client_type = Column(Enum(ClientType), default=ClientType.PRIVATE)
    name = Column(String(255))
    redirect_uris = Column(String(255), default=None)
    user = relationship(User, back_populates='oauth2_applications')
    user_id = Column(Integer, ForeignKey('user.id'))

    oauth2_access_tokens = relationship('Oauth2AccessToken', back_populates='oauth2_application')

    @classmethod
    @safe_db_query
    def query_client(self, api_key: str):
        return self.query.filter(
            Oauth2Application.client_id == api_key,
        ).first()


class Oauth2AccessToken(BaseModel):
    expires = Column(DateTime(timezone=True))
    oauth2_application = relationship(Oauth2Application, back_populates='oauth2_access_tokens')
    oauth2_application_id = Column(Integer, ForeignKey('oauth2_application.id'))
    token = Column(String(255), index=True, unique=True)
    user = relationship(User, back_populates='oauth2_access_tokens')
    user_id = Column(Integer, ForeignKey('user.id'))

    def is_valid(self) -> bool:
        return self.token and \
            self.expires and \
            self.expires >= datetime.utcnow().replace(tzinfo=self.expires.tzinfo)



from mage_ai.api.errors import ApiError
from mage_ai.api.resources.GenericResource import GenericResource
from mage_ai.orchestration.db import safe_db_query
from mage_ai.data_preparation.models.variable import VariableType
from mage_ai.data_preparation.repo_manager import get_variables_dir
from mage_ai.data_preparation.variable_manager import (
    VariableManager,
    delete_global_variable,
    get_global_variables,
    set_global_variable,
)
from typing import Dict


def get_variable_value(
    variable_manager: VariableManager,
    pipeline_uuid: str,
    block_uuid: str,
    variable_uuid: str,
) -> Dict:
    variable = variable_manager.get_variable_object(
        pipeline_uuid,
        block_uuid,
        variable_uuid,
    )
    if variable.variable_type in [VariableType.DATAFRAME, VariableType.GEO_DATAFRAME]:
        value = 'DataFrame'
        variable_type = 'pandas.DataFrame'
    else:
        value = variable.read_data(sample=True)
        variable_type = str(type(value))
    return dict(
        uuid=variable_uuid,
        type=variable_type,
        value=value,
    )


class VariableResource(GenericResource):
    @classmethod
    @safe_db_query
    def collection(self, query, meta, user, **kwargs):
        pipeline_uuid = kwargs['parent_model'].uuid

        # Get global variables from project's path
        variable_manager = VariableManager(variables_dir=get_variables_dir())
        variables_dict = variable_manager.get_variables_by_pipeline(pipeline_uuid)
        global_variables = [
            dict(
                uuid=uuid,
                type=str(type(value)),
                value=value
            )
            for uuid, value in get_global_variables(pipeline_uuid).items()
        ]
        global_variables_arr = [
            dict(
                block=dict(uuid='global'),
                pipeline=dict(uuid=pipeline_uuid),
                variables=global_variables,
            )
        ]
        variables = [
            dict(
                block=dict(uuid=uuid),
                pipeline=dict(uuid=pipeline_uuid),
                variables=[
                    get_variable_value(variable_manager, pipeline_uuid, uuid, var) for var in arr
                    # Not return printed outputs
                    if var == 'df' or var.startswith('output')
                  ],
            )
            for uuid, arr in variables_dict.items() if uuid != 'global'
        ] + global_variables_arr

        return self.build_result_set(
            variables,
            user,
            **kwargs,
        )

    @classmethod
    @safe_db_query
    def create(self, payload: Dict, user, **kwargs) -> 'VariableResource':
        pipeline_uuid = kwargs['parent_model'].uuid

        error = ApiError.RESOURCE_INVALID.copy()

        variable_uuid = payload.get('name')
        if not variable_uuid.isidentifier():
            error.update(message=f'Invalid variable name syntax for variable name {variable_uuid}.')
            raise ApiError(error)

        variable_value = payload.get('value')
        if variable_value is None:
            error.update(message=f'Value is empty for variable name {variable_uuid}.')
            raise ApiError(error)

        set_global_variable(
            pipeline_uuid,
            variable_uuid,
            variable_value,
        )

        global_variables = get_global_variables(pipeline_uuid)

        return self(dict(
            block=dict(uuid='global'),
            name=variable_uuid,
            pipeline=dict(uuid=pipeline_uuid),
            value=variable_value,
            variables=list(global_variables.keys()),
        ), user, **kwargs)

    @classmethod
    @safe_db_query
    def member(self, pk, user, **kwargs):
        return self(dict(name=pk), user, **kwargs)

    @safe_db_query
    def update(self, payload, **kwargs):
        pipeline_uuid = kwargs['parent_model'].uuid

        error = ApiError.RESOURCE_INVALID.copy()

        variable_uuid = payload.get('name')
        if not variable_uuid.isidentifier():
            error.update(message=f'Invalid variable name syntax for variable name {self.name}.')
            raise ApiError(error)

        variable_value = payload.get('value')
        if variable_value is None:
            error.update(message=f'Value is empty for variable name {self.name}.')
            raise ApiError(error)

        set_global_variable(
            pipeline_uuid,
            variable_uuid,
            variable_value,
        )

        global_variables = get_global_variables(pipeline_uuid)

        self.model.update(dict(
            block=dict(uuid='global'),
            name=variable_uuid,
            pipeline=dict(uuid=pipeline_uuid),
            value=variable_value,
            variables=list(global_variables.keys()),
        ))

        return self

    @safe_db_query
    def delete(self, **kwargs):
        pipeline_uuid = kwargs['parent_model'].uuid
        delete_global_variable(pipeline_uuid, self.name)
        return self



from mage_ai.shared.io import safe_write, safe_write_async
from mage_ai.tests.base_test import TestCase
import asyncio
import os


class IOTests(TestCase):
    def test_safe_write(self):
        file_path = 'test_file_write'
        with open(file_path, 'w') as fp:
            fp.write('MESSAGE1')

        safe_write(file_path, 'MESSAGE2')
        with open(file_path, 'r') as fp:
            self.assertEqual(fp.read(), 'MESSAGE2')

        def write_func(fp, content):
            raise Exception('Fail to write')

        with self.assertRaises(Exception) as err:
            safe_write(file_path, 'MESSAGE3', write_func=write_func)
            self.assertTrue('Fail to write' in str(err.exception))
        with open(file_path, 'r') as fp:
            self.assertEqual(fp.read(), 'MESSAGE2')
        os.remove(file_path)

    def test_safe_write_sync(self):
        file_path = 'test_file_write_2'
        with open(file_path, 'w') as fp:
            fp.write('MESSAGE1')

        asyncio.run(safe_write_async(file_path, 'MESSAGE2'))
        with open(file_path, 'r') as fp:
            self.assertEqual(fp.read(), 'MESSAGE2')

        async def write_func_async(fp, content):
            raise Exception('Fail to write async')

        with self.assertRaises(Exception) as err:
            asyncio.run(safe_write_async(file_path, 'MESSAGE3', write_func=write_func_async))
            self.assertTrue('Fail to write async' in str(err.exception))
        with open(file_path, 'r') as fp:
            self.assertEqual(fp.read(), 'MESSAGE2')
        os.remove(file_path)



from mage_ai.api.presenters.BasePresenter import BasePresenter


class AutocompleteItemPresenter(BasePresenter):
    default_attributes = [
        'classes',
        'constants',
        'files',
        'functions',
        'group',
        'id',
        'imports',
        'methods_for_class',
    ]



from mage_ai.api.resources.GenericResource import GenericResource
from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.orchestration.db import safe_db_query
from mage_ai.server.constants import VERSION
import aiohttp


class ProjectResource(GenericResource):
    @classmethod
    @safe_db_query
    async def collection(self, query, meta, user, **kwargs):
        try:
            async with aiohttp.ClientSession() as session:
                async with session.get(
                    'https://pypi.org/pypi/mage-ai/json',
                    timeout=3,
                ) as response:
                    response_json = await response.json()
                    latest_version = response_json.get('info', {}).get('version', None)
        except Exception:
            latest_version = VERSION

        parts = get_repo_path().split('/')

        collection = [
            dict(
                latest_version=latest_version,
                name=parts[-1],
                version=VERSION,
            ),
        ]

        return self.build_result_set(
            collection,
            user,
            **kwargs,
        )



from google.cloud.bigquery import Client, LoadJobConfig, WriteDisposition
from google.oauth2 import service_account
from mage_ai.io.base import BaseSQLDatabase, ExportWritePolicy, QUERY_ROW_LIMIT
from mage_ai.io.config import BaseConfigLoader, ConfigKey
from mage_ai.io.export_utils import infer_dtypes
from mage_ai.shared.utils import (
    convert_pandas_dtype_to_python_type,
    convert_python_type_to_bigquery_type,
)
from pandas import DataFrame
from typing import Dict, List, Mapping, Union


class BigQuery(BaseSQLDatabase):
    """
    Handles data transfer betwee a BigQuery data warehouse and the Mage app.
    """

    def __init__(self, **kwargs) -> None:
        """
        Initializes settings for connecting to a BigQuery warehouse.

        To authenticate (and authorize) access to a BigQuery warehouse, credentials must be
        provided.
        Below are the different ways in which the BigQuery data loader can access those
        credentials:
        - Define the `GOOGLE_APPLICATION_CREDENTIALS` environment variable to point to a
          service account key. In this case no other no other parameters need to be
          specified.
        - Manually pass in the `google.oauth2.service_account.Credentials` object with the
        keyword argument `credentials`
        - Manually pass in the path to the credentials with the keyword argument
        `path_to_credentials`.
        - Manually define the service key key-value set to use (such as a dictionary containing
        the same parameters as a service key) with the keyword argument `credentials_mapping`

        All keyword arguments except for `path_to_credentials` and `credentials_mapping` will be
        passed to the Google BigQuery client, accepting all other configuration settings there.
        """
        if kwargs.get('verbose') is not None:
            kwargs.pop('verbose')
        super().__init__(verbose=kwargs.get('verbose', True))

        credentials = kwargs.get('credentials')
        if credentials is None:
            if 'credentials_mapping' in kwargs:
                mapping_obj = kwargs.pop('credentials_mapping')
                if mapping_obj is not None:
                    credentials = service_account.Credentials.from_service_account_info(mapping_obj)
            if 'path_to_credentials' in kwargs:
                path = kwargs.pop('path_to_credentials')
                if path is not None:
                    credentials = service_account.Credentials.from_service_account_file(path)
            if 'credentials' in kwargs:
                kwargs.pop('credentials')
        with self.printer.print_msg('Connecting to BigQuery warehouse'):
            self.client = Client(credentials=credentials, **kwargs)

    def default_database(self) -> str:
        return self.client.project

    def get_column_types(self, schema: str, table_name: str) -> Dict:
        results = self.client.query(f"""
SELECT
    column_name
    , data_type
FROM {schema}.INFORMATION_SCHEMA.COLUMNS
WHERE TABLE_NAME = '{table_name}'
""")

        column_types = {}
        for col, col_type in results:
            column_types[col.lower()] = col_type

        return column_types

    def alter_table(
        self,
        df: DataFrame,
        table_id: str,
        database: Union[str, None] = None,
        verbose: bool = True,
    ):
        def __process(database):
            parts = table_id.split('.')
            if len(parts) == 2:
                schema, table_name = parts
            elif len(parts) == 3:
                database, schema, table_name = parts

            df_existing = self.client.query(f"""
    SELECT 1
    FROM `{database}.{schema}.__TABLES_SUMMARY__`
    WHERE table_id = '{table_name}'
    """).to_dataframe()

            full_table_name = f'{database}.{schema}.{table_name}'

            table_doesnt_exist = df_existing.empty

            if table_doesnt_exist:
                raise ValueError(f'Table \'{table_id}\' doesn\'t exist.')

            current_columns = list(self.get_column_types(schema, table_name).keys())

            columns = []
            if type(df) is DataFrame:
                columns = df.columns
            elif type(df) is dict:
                columns = df.keys()

            new_columns = [c for c in columns if c.lower() not in current_columns]
            if not new_columns:
                return
            dtypes = infer_dtypes(df)
            dtypes = \
                {k: convert_python_type_to_bigquery_type(convert_pandas_dtype_to_python_type(v))
                 for k, v in dtypes.items()}
            columns_and_types = [
                f"ADD COLUMN {self._clean_column_name(col)} {dtypes[col]}" for col
                in new_columns
            ]
            # TODO: support alter column type and drop columns
            alter_table_command = f"ALTER TABLE {full_table_name} {', '.join(columns_and_types)}"
            self.client.query(alter_table_command)
        if verbose:
            with self.printer.print_msg(f'Altering table \'{table_id}\''):
                __process(database=database)
        else:
            __process(database=database)

    def load(
        self,
        query_string: str,
        limit: int = QUERY_ROW_LIMIT,
        display_query: Union[str, None] = None,
        verbose: bool = True,
        **kwargs,
    ) -> DataFrame:
        """
        Loads data from BigQuery into a Pandas data frame based on the query given.
        This will fail if the query returns no data from the database. When a select query
        is provided, this function will load at maximum 10,000,000 rows of data. To operate on more
        data, consider performing data transformations in warehouse.

        Args:
            query_string (str): Query to fetch a table or subset of a table.
            limit (int, Optional): The number of rows to limit the loaded dataframe to. Defaults to
                                    10,000,000.
            **kwargs: Additional arguments to pass to query, such as query configurations

        Returns:
            DataFrame: Data frame associated with the given query.
        """

        print_message = 'Loading data'
        if verbose:
            print_message += ' with query'

            if display_query:
                for line in display_query.split('\n'):
                    print_message += f'\n{line}'
            else:
                print_message += f'\n{query_string}'

        query_string = self._clean_query(query_string)

        with self.printer.print_msg(print_message):
            return self.client.query(
                self._enforce_limit(query_string, limit), *kwargs
            ).to_dataframe()

    def export(
        self,
        df: DataFrame,
        table_id: str,
        database: Union[str, None] = None,
        if_exists: str = 'replace',
        query_string: Union[str, None] = None,
        verbose: bool = True,
        **configuration_params,
    ) -> None:
        """
        Exports a data frame to a Google BigQuery warehouse.  If table doesn't
        exist, the table is automatically created.

        Args:
            df (DataFrame): Data frame to export
            table_id (str): ID of the table to export the data frame to. If of the format
            `"your-project.your_dataset.your_table_name"`. If this table exists,
            the table schema must match the data frame schema. If this table doesn't exist,
            the table schema is automatically inferred.
            if_exists (str): Specifies export policy if table exists. Either
                - `'fail'`: throw an error.
                - `'replace'`: drops existing table and creates new table of same name.
                - `'append'`: appends data frame to existing table. In this case the schema must
                                match the original table.
            Defaults to `'replace'`. If `write_disposition` is specified as a keyword argument,
            this parameter is ignored (as both define the same functionality).
            **configuration_params: Configuration parameters for export job
        """

        if type(df) is dict:
            df = DataFrame([df])
        elif type(df) is list:
            df = DataFrame(df)

        def __process(database: Union[str, None]):
            if query_string:
                parts = table_id.split('.')
                if len(parts) == 2:
                    schema, table_name = parts
                elif len(parts) == 3:
                    database, schema, table_name = parts

                df_existing = self.client.query(f"""
SELECT 1
FROM `{database}.{schema}.__TABLES_SUMMARY__`
WHERE table_id = '{table_name}'
""").to_dataframe()

                full_table_name = f'{database}.{schema}.{table_name}'

                table_doesnt_exist = df_existing.empty

                if ExportWritePolicy.FAIL == if_exists and not table_doesnt_exist:
                    raise ValueError(
                        f'Table \'{full_table_name}\' already exists in database.',
                    )

                if ExportWritePolicy.REPLACE == if_exists:
                    self.client.query(f'DROP TABLE {full_table_name}')
                    command = f'CREATE TABLE {table_id} AS'
                elif table_doesnt_exist:
                    command = f'CREATE TABLE {table_id} AS'
                else:
                    command = f'INSERT INTO {table_id}'

                sql = f"""
{command}
{query_string}
"""
                self.client.query(sql)

            else:
                config = LoadJobConfig(**configuration_params)
                if 'write_disposition' not in configuration_params:
                    if if_exists == 'replace':
                        config.write_disposition = WriteDisposition.WRITE_TRUNCATE
                    elif if_exists == 'append':
                        config.write_disposition = WriteDisposition.WRITE_APPEND
                    elif if_exists == 'fail':
                        config.write_disposition = WriteDisposition.WRITE_EMPTY
                    else:
                        raise ValueError(
                            f'Invalid policy specified for handling existence of '
                            f'table: \'{if_exists}\''
                        )
                parts = table_id.split('.')
                if len(parts) == 2:
                    schema = parts[0]
                    table_name = parts[1]
                elif len(parts) == 3:
                    schema = parts[1]
                    table_name = parts[2]

                self.client.create_dataset(dataset=schema, exists_ok=True)

                column_types = self.get_column_types(schema, table_name)

                for col in df.columns:
                    col_type = column_types.get(col)
                    if not col_type:
                        continue

                    null_rows = df[col].isnull()
                    if col_type.startswith('ARRAY<STRUCT'):
                        df.loc[null_rows, col] = df.loc[null_rows, col].apply(lambda x: [{}])
                    elif col_type.startswith('ARRAY'):
                        df.loc[null_rows, col] = df.loc[null_rows, col].apply(lambda x: [])
                    elif col_type.startswith('STRUCT'):
                        df.loc[null_rows, col] = df.loc[null_rows, col].apply(lambda x: {})

                # Clean column names
                if type(df) is DataFrame:
                    df.columns = df.columns.str.replace(' ', '_')

                self.client.load_table_from_dataframe(df, table_id, job_config=config).result()

        if verbose:
            with self.printer.print_msg(f'Exporting data to table \'{table_id}\''):
                __process(database=database)
        else:
            __process(database=database)

    def execute(self, query_string: str, **kwargs) -> None:
        """
        Sends query to the connected BigQuery warehouse.

        Args:
            query_string (str): Query to execute on the BigQuery warehouse.
            **kwargs: Additional arguments to pass to query, such as query configurations
        """
        with self.printer.print_msg(f'Executing query \'{query_string}\''):
            query_string = self._clean_query(query_string)
            self.client.query(query_string, **kwargs)

    def execute_queries(
        self,
        queries: List[str],
        query_variables: List[Dict] = None,
        fetch_query_at_indexes: List[bool] = None,
        **kwargs,
    ) -> List:
        results = []

        for idx, query in enumerate(queries):
            variables = query_variables[idx] \
                        if query_variables and idx < len(query_variables) \
                        else {}
            query = self._clean_query(query)
            result = self.client.query(query, **variables)

            if fetch_query_at_indexes and idx < len(fetch_query_at_indexes) and \
                    fetch_query_at_indexes[idx]:
                result = result.to_dataframe()

            results.append(result)

        return results

    @classmethod
    def with_config(cls, config: BaseConfigLoader, **kwargs) -> 'BigQuery':
        """
        Initializes BigQuery client from configuration loader

        Args:
            config (BaseConfigLoader): Configuration loader object
        """
        if ConfigKey.GOOGLE_SERVICE_ACC_KEY in config:
            kwargs['credentials_mapping'] = config[ConfigKey.GOOGLE_SERVICE_ACC_KEY]
        elif ConfigKey.GOOGLE_SERVICE_ACC_KEY_FILEPATH in config:
            kwargs['path_to_credentials'] = config[ConfigKey.GOOGLE_SERVICE_ACC_KEY_FILEPATH]
        else:
            raise ValueError(
                'No valid configuration settings found for Google BigQuery. You must specify '
                'either your service account key or the filepath to your service account key.'
            )

        if ConfigKey.GOOGLE_LOCATION in config:
            kwargs['location'] = config[ConfigKey.GOOGLE_LOCATION]

        return cls(**kwargs)

    @classmethod
    def with_credentials_file(cls, path_to_credentials: str, **kwargs) -> 'BigQuery':
        """
        Constructs BigQuery data loader using the file containing the service account key.

        Args:
            path_to_credentials (str): Path to the credentials file.

        Returns:
            BigQuery: BigQuery data loader
        """
        return cls(path_to_credentials=path_to_credentials, **kwargs)

    @classmethod
    def with_credentials_object(cls, credentials: Mapping[str, str], **kwargs) -> 'BigQuery':
        """
        Constructs BigQuery data loader using manually specified authentication credentials object.

        Args:
            credentials (Mapping[str, str]): Credentials object. Must contain all the OAuth
                information necessary to authenticate and authorize BigQuery access. This should
                resemble a service account key.

        Returns:
            BigQuery: BigQuery data loader
        """
        return cls(credentials_mapping=credentials, **kwargs)



from mage_ai.data_preparation.models.file import File
from mage_ai.data_preparation.storage.base_storage import BaseStorage
from mage_ai.shared.parsers import encode_complex
from typing import Dict, List
import aiofiles
import json
import os
import pandas as pd
import shutil
import simplejson


class LocalStorage(BaseStorage):
    def isdir(self, path: str) -> bool:
        return os.path.isdir(path)

    def listdir(self, path: str, suffix: str = None) -> List[str]:
        if not os.path.exists(path):
            return []
        paths = os.listdir(path)
        if suffix is not None:
            paths = [p for p in paths if p.endswith(suffix)]
        return paths

    def makedirs(self, path: str, **kwargs) -> None:
        os.makedirs(path, exist_ok=True)

    def path_exists(self, path: str) -> bool:
        return os.path.exists(path)

    def remove(self, path: str) -> None:
        os.remove(path)

    def remove_dir(self, path: str) -> None:
        shutil.rmtree(path, ignore_errors=True)

    def read_json_file(self, file_path: str, default_value={}) -> Dict:
        if not self.path_exists(file_path):
            return default_value
        with open(file_path) as file:
            try:
                return json.load(file)
            except Exception:
                return default_value

    async def read_json_file_async(self, file_path: str, default_value={}) -> Dict:
        if not self.path_exists(file_path):
            return default_value
        async with aiofiles.open(file_path, mode='r') as file:
            try:
                return json.loads(await file.read())
            except Exception:
                return default_value

    def write_json_file(self, file_path: str, data) -> None:
        with open(file_path, 'w') as file:
            simplejson.dump(
                data,
                file,
                default=encode_complex,
                ignore_nan=True,
            )

    async def write_json_file_async(self, file_path: str, data) -> None:
        async with aiofiles.open(file_path, mode='w') as file:
            fcontent = simplejson.dumps(
                data,
                default=encode_complex,
                ignore_nan=True,
            )
            await file.write(fcontent)

    def read_parquet(self, file_path: str, **kwargs) -> pd.DataFrame:
        return pd.read_parquet(file_path, engine='pyarrow')

    def write_csv(self, df: pd.DataFrame, file_path: str) -> None:
        File.create_parent_directories(file_path)
        df.to_csv(file_path, index=False)

    def write_parquet(self, df: pd.DataFrame, file_path: str) -> None:
        File.create_parent_directories(file_path)
        df.to_parquet(file_path)



from mage_ai.shared.retry import retry
from mage_ai.tests.base_test import TestCase
from unittest.mock import call, patch


class RetryTests(TestCase):
    @patch('time.sleep')
    def test_retry(self, mock_sleep):
        @retry(retries=3, max_delay=40)
        def test_func():
            raise Exception('error')
            return

        with self.assertRaises(Exception):
            test_func()
            mock_sleep.assert_has_calls([call(5), call(10), call(20)])



from mage_ai.api.oauth_scope import OauthScope
from mage_ai.api.operations import constants
from mage_ai.api.policies.BasePolicy import BasePolicy
from mage_ai.api.presenters.BackfillPresenter import BackfillPresenter


class BackfillPolicy(BasePolicy):
    pass


BackfillPolicy.allow_actions([
    constants.DETAIL,
    constants.LIST,
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], condition=lambda policy: policy.has_at_least_viewer_role())

BackfillPolicy.allow_actions([
    constants.CREATE,
    constants.UPDATE,
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], condition=lambda policy: policy.has_at_least_editor_role_and_edit_access())

BackfillPolicy.allow_read([
    'pipeline_run_dates',
    'total_run_count',
] + BackfillPresenter.default_attributes, scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.DETAIL,
    constants.LIST,
], condition=lambda policy: policy.has_at_least_viewer_role())

BackfillPolicy.allow_read([] + BackfillPresenter.default_attributes, scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.CREATE,
    constants.UPDATE,
], condition=lambda policy: policy.has_at_least_editor_role_and_edit_access())

BackfillPolicy.allow_write([
    'name',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.CREATE,
], condition=lambda policy: policy.has_at_least_editor_role_and_edit_access())

BackfillPolicy.allow_write([
    'block_uuid',
    'end_datetime',
    'interval_type',
    'interval_units',
    'name',
    'start_datetime',
    'status',
    'variables',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.UPDATE,
], condition=lambda policy: policy.has_at_least_editor_role_and_edit_access())

BackfillPolicy.allow_query([
    'include_preview_runs',
    'include_run_count',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.DETAIL,
], condition=lambda policy: policy.has_at_least_viewer_role())

BackfillPolicy.allow_query([
    'pipeline_uuid',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.LIST,
], condition=lambda policy: policy.has_at_least_viewer_role())



from .constants import (
    AggregationFunction,
    VARIABLE_NAME_X,
    VARIABLE_NAME_Y,
)
from mage_ai.shared.parsers import encode_complex
import numpy as np
import pandas as pd


def clean_series(series, column_type=None, dropna=True):
    series_cleaned = series.map(
        lambda x: x if (not isinstance(x, str) or (len(x) > 0 and not x.isspace())) else np.nan,
    )
    if dropna:
        series_cleaned = series_cleaned.dropna()

    if column_type is int:
        try:
            series_cleaned = series_cleaned.astype(float).astype(int)
        except ValueError:
            series_cleaned = series_cleaned.astype(float)
    elif column_type is float:
        series_cleaned = series_cleaned.astype(float)

    return series_cleaned


def convert_to_list(arr, limit=None):
    if type(arr) in [pd.Index, pd.RangeIndex, pd.Series]:
        return arr[:limit].tolist()
    elif type(arr) is pd.DataFrame:
        return arr[:limit].to_numpy().tolist()
    elif type(arr) is np.ndarray:
        return arr[:limit].tolist()
    elif type(arr) is list:
        return [convert_to_list(arr2) for arr2 in arr]

    return arr


def encode_values_in_list(arr):
    return [encode_complex(v) for v in arr]


def build_metric_name(metric):
    aggregation = metric['aggregation']
    column = metric['column']
    return f'{aggregation}({column})'


def calculate_metric_for_series(series, aggregation):
    series = clean_series(series)
    value = 0

    if AggregationFunction.AVERAGE == aggregation:
        count = len(series)
        if count > 0:
            value = sum(series) / count
    elif AggregationFunction.COUNT == aggregation:
        value = len(series)
    elif AggregationFunction.COUNT_DISTINCT == aggregation:
        value = len(series.unique())
    elif AggregationFunction.MAX == aggregation:
        value = max(series)
    elif AggregationFunction.MEDIAN == aggregation:
        value = sorted(series)[int(len(series) / 2)]
    elif AggregationFunction.MIN == aggregation:
        value = min(series)
    elif AggregationFunction.MODE == aggregation:
        value = sorted(
            series.value_counts().items(),
            key=lambda t: t[1],
            reverse=True,
        )[0][0]
    elif AggregationFunction.SUM == aggregation:
        value = sum(series)

    return value


def calculate_metrics_for_group(metrics, group):
    values = {}

    for metric in metrics:
        aggregation = metric['aggregation']
        column = metric['column']
        series = group[column]

        values[build_metric_name(metric)] = calculate_metric_for_series(series, aggregation)

    return values


def build_x_y(df, group_by_columns, metrics):
    data = {}
    groups = df.groupby(group_by_columns)
    data[VARIABLE_NAME_X] = list(groups.groups.keys())

    metrics_per_group = groups.apply(
        lambda group: calculate_metrics_for_group(metrics, group),
    ).values

    y_values = []
    for idx, metric in enumerate(metrics):
        y_values.append([g[build_metric_name(metric)] for g in metrics_per_group])

    data[VARIABLE_NAME_Y] = y_values

    return data



class ApiContext():
    # DANGER: do not allow this class to instantiate with any arguments.
    # For some reason and some unknown location previously, this class was being instantiated
    # with arguments that included some sort of stale data structure that was causing
    # API permission checking issues for every other request because the ResultSet was being
    # persisted somehow between requests.
    def __init__(self):
        self.data = {}



from abc import ABC, abstractmethod
from typing import Dict, List


class BaseStorage(ABC):
    @abstractmethod
    def isdir(self, path: str) -> bool:
        """
        Check whether a path is a directory.
        """
        pass

    @abstractmethod
    def listdir(self, path: str, suffix: str = None) -> List[str]:
        """
        Get a list of files and directories in the specified directory.
        """
        pass

    @abstractmethod
    def makedirs(self, path: str, **kwargs) -> None:
        """
        Create new directories.
        """
        pass

    @abstractmethod
    def path_exists(self, path: str) -> bool:
        """
        Check whether a path exists.
        """
        pass

    @abstractmethod
    def remove(self, path: str) -> None:
        """
        Remove a file with file path.
        """
        pass

    @abstractmethod
    def remove_dir(self, path: str) -> None:
        """
        Remove a directory with directory path.
        """
        pass

    @abstractmethod
    def read_json_file(self, file_path: str, default_value={}) -> Dict:
        """
        Read json from a file with file path.
        """
        pass

    @abstractmethod
    async def read_json_file_async(self, file_path: str, default_value={}) -> Dict:
        """
        Read json from a file with file path asynchronously.
        """
        pass

    @abstractmethod
    def write_json_file(self, file_path: str, data) -> None:
        """
        Write json to a file with file path.
        """
        pass

    @abstractmethod
    async def write_json_file_async(self, file_path: str, data) -> None:
        """
        Write json to a file with file path asynchronously.
        """
        pass



"""Add event matcher model.

Revision ID: 3fafd217efa7
Revises: 053ee2c10d85
Create Date: 2022-09-01 11:48:03.057401

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '3fafd217efa7'
down_revision = '053ee2c10d85'
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('event_matcher',
    sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
    sa.Column('created_at', sa.DateTime(timezone=True), server_default=sa.text('(CURRENT_TIMESTAMP)'), nullable=True),
    sa.Column('updated_at', sa.DateTime(timezone=True), server_default=sa.text('(CURRENT_TIMESTAMP)'), nullable=True),
    sa.Column('event_type', sa.Enum('AWS_EVENT', name='eventtype'), nullable=True),
    sa.Column('name', sa.String(length=255), nullable=True),
    sa.Column('pattern', sa.JSON(), nullable=True),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_table('pipeline_schedule_event_matcher_association',
    sa.Column('pipeline_schedule_id', sa.Integer(), nullable=True),
    sa.Column('event_matcher_id', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['event_matcher_id'], ['event_matcher.id'], ),
    sa.ForeignKeyConstraint(['pipeline_schedule_id'], ['pipeline_schedule.id'], )
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('pipeline_schedule_event_matcher_association')
    op.drop_table('event_matcher')
    # ### end Alembic commands ###



from dataclasses import dataclass
from mage_ai.shared.config import BaseConfig
from mage_ai.streaming.sources.base import BaseSource
from typing import Callable
import boto3
import json

DEFAULT_BATCH_SIZE = 100


@dataclass
class KinesisConfig(BaseConfig):
    stream_name: str
    batch_size: int = DEFAULT_BATCH_SIZE


class KinesisSource(BaseSource):
    config_class = KinesisConfig

    def init_client(self):
        self._print('Start initializing consumer.')
        # Initialize kafka consumer
        self.kinesis_client = boto3.client('kinesis')
        self.details = self.kinesis_client.describe_stream(
            StreamName=self.config.stream_name,
        )['StreamDescription']
        self._print('Finish initializing consumer.')

    def read(self, handler: Callable):
        pass

    def batch_read(self, handler: Callable):
        self._print('Start consuming messages.')
        if self.config.batch_size > 0:
            batch_size = self.config.batch_size
        else:
            batch_size = DEFAULT_BATCH_SIZE

        try:
            shard_iterators = dict()
            for shard_id in self.details['Shards']:
                shard_id = shard_id['ShardId']
                iterator_kwargs = dict(
                    ShardIteratorType='LATEST',
                )
                if self.checkpoint is not None and self.checkpoint.get(shard_id):
                    iterator_kwargs = dict(
                        ShardIteratorType='AFTER_SEQUENCE_NUMBER',
                        StartingSequenceNumber=self.checkpoint.get(shard_id),
                    )
                self._print(f'Shard {shard_id} args: {iterator_kwargs}')
                shard_iterator = self.kinesis_client.get_shard_iterator(
                    StreamName=self.config.stream_name,
                    ShardId=shard_id,
                    **iterator_kwargs,
                )
                shard_iterators[shard_id] = shard_iterator['ShardIterator']
            self._print(f'Consuming messages from shards: {list(shard_iterators.keys())}')
            closed_streams = set()
            while True:
                for shard_id, shard_iter in shard_iterators.items():
                    if shard_id in closed_streams:
                        continue
                    response = self.kinesis_client.get_records(
                        ShardIterator=shard_iter,
                        Limit=batch_size,
                    )
                    if 'NextShardIterator' not in response:
                        self._print(f'Shards {shard_id} is closed.')
                        closed_streams.add(shard_id)
                        continue
                    shard_iterators[shard_id] = response['NextShardIterator']
                    records = response['Records']
                    if records:
                        self._print(f'Got {len(records)} records from shard {shard_id}. '
                                    f'Sample: {records[0]}')
                        handler([json.loads(r['Data'].decode('utf-8')) for r in records])
                        max_seq_number = max(r['SequenceNumber'] for r in records)
                        self.update_checkpoint(shard_id, max_seq_number)
                if len(closed_streams) == len(shard_iterators):
                    self._print(f'All shards {list(shard_iterators.keys())} are closed.')
                    break
        except Exception:
            self._print(f'Couldn\'t get records from stream {self.config.stream_name}.')
            raise

    def update_checkpoint(self, shard_id, sequence_number):
        if self.checkpoint is None:
            self.checkpoint = {}
        if not self.checkpoint.get(shard_id):
            self.checkpoint[shard_id] = sequence_number
        elif sequence_number > self.checkpoint.get(shard_id):
            self.checkpoint[shard_id] = sequence_number
        super().update_checkpoint()



from mage_ai.data_cleaner.transformer_actions.constants import VariableType
import re


def interpolate(text, key, variable_data):
    """
    text:
        string to operate on
    key:
        key to search within text
    variable_data:
        dictionary containing data used to interpolate
    """

    regex_replacement = key
    if variable_data['type'] == VariableType.FEATURE:
        regex_replacement = variable_data[VariableType.FEATURE]['uuid']
    elif variable_data['type'] == VariableType.FEATURE_SET_VERSION:
        regex_replacement = \
            variable_data[VariableType.FEATURE_SET_VERSION][VariableType.FEATURE_SET]['uuid']

    regex_pattern = re.compile(
        r'\%__BRACKETS_START__{}__BRACKETS_END__'
        .format(key)
        .replace('__BRACKETS_START__', r'\{')
        .replace('__BRACKETS_END__', r'\}')
    )

    return re.sub(regex_pattern, regex_replacement, str(text))


def replace_true_false(action_code):
    regex_pattern_true = re.compile(' true')
    regex_pattern_false = re.compile(' false')

    return re.sub(
        regex_pattern_true,
        ' True',
        re.sub(regex_pattern_false, ' False', action_code),
    )



from dataclasses import dataclass
from kafka import KafkaConsumer
from mage_ai.shared.config import BaseConfig
from mage_ai.streaming.constants import DEFAULT_BATCH_SIZE
from mage_ai.streaming.sources.base import BaseSource
from mage_ai.streaming.sources.shared import SerializationMethod, SerDeConfig
from enum import Enum
from typing import Callable, Dict
import importlib
import json
import time


class SecurityProtocol(str, Enum):
    SASL_SSL = 'SASL_SSL'
    SSL = 'SSL'


@dataclass
class SASLConfig:
    mechanism: str = 'PLAIN'
    username: str = None
    password: str = None


@dataclass
class SSLConfig:
    cafile: str = None
    certfile: str = None
    keyfile: str = None
    password: str = None
    check_hostname: bool = False


@dataclass
class KafkaConfig(BaseConfig):
    bootstrap_server: str
    consumer_group: str
    topic: str
    api_version: str = '0.10.2'
    batch_size: int = DEFAULT_BATCH_SIZE
    security_protocol: SecurityProtocol = None
    ssl_config: SSLConfig = None
    sasl_config: SASLConfig = None
    serde_config: SerDeConfig = None

    @classmethod
    def parse_config(self, config: Dict) -> Dict:
        ssl_config = config.get('ssl_config')
        sasl_config = config.get('sasl_config')
        serde_config = config.get('serde_config')
        if ssl_config and type(ssl_config) is dict:
            config['ssl_config'] = SSLConfig(**ssl_config)
        if sasl_config and type(sasl_config) is dict:
            config['sasl_config'] = SASLConfig(**sasl_config)
        if serde_config and type(serde_config) is dict:
            config['serde_config'] = SerDeConfig(**serde_config)
        return config


class KafkaSource(BaseSource):
    config_class = KafkaConfig

    def init_client(self):
        self._print('Start initializing consumer.')
        # Initialize kafka consumer
        consumer_kwargs = dict(
            group_id=self.config.consumer_group,
            bootstrap_servers=self.config.bootstrap_server,
            api_version=self.config.api_version,
            enable_auto_commit=True,
        )
        if self.config.security_protocol == SecurityProtocol.SSL:
            consumer_kwargs['security_protocol'] = SecurityProtocol.SSL
            consumer_kwargs['ssl_cafile'] = self.config.ssl_config.cafile
            consumer_kwargs['ssl_certfile'] = self.config.ssl_config.certfile
            consumer_kwargs['ssl_keyfile'] = self.config.ssl_config.keyfile
            consumer_kwargs['ssl_password'] = self.config.ssl_config.password
            consumer_kwargs['ssl_check_hostname'] = self.config.ssl_config.check_hostname
        elif self.config.security_protocol == SecurityProtocol.SASL_SSL:
            consumer_kwargs['security_protocol'] = SecurityProtocol.SASL_SSL
            consumer_kwargs['sasl_mechanism'] = self.config.sasl_config.mechanism
            consumer_kwargs['sasl_plain_username'] = self.config.sasl_config.username
            consumer_kwargs['sasl_plain_password'] = self.config.sasl_config.password

        self.consumer = KafkaConsumer(
            self.config.topic,
            **consumer_kwargs
        )
        self._print('Finish initializing consumer.')

        self.schema_class = None
        if self.config.serde_config is not None and \
                self.config.serde_config.serialization_method == SerializationMethod.PROTOBUF:
            schema_classpath = self.config.serde_config.schema_classpath
            if schema_classpath is None:
                return
            self._print(f'Loading message schema from {schema_classpath}')
            parts = schema_classpath.split('.')
            if len(parts) >= 2:
                class_name = parts[-1]
                libpath = '.'.join(parts[:-1])
                self.schema_class = getattr(
                    importlib.import_module(libpath),
                    class_name,
                )

    def read(self, handler: Callable):
        self._print('Start consuming messages.')
        for message in self.consumer:
            self.__print_message(message)
            data = self.__deserialize_message(message.value)
            handler(data)

    async def read_async(self, handler: Callable):
        self._print('Start consuming messages asynchronously.')
        for message in self.consumer:
            self.__print_message(message)
            data = self.__deserialize_message(message.value)
            await handler(data)

    def batch_read(self, handler: Callable):
        self._print('Start consuming messages.')
        if self.config.batch_size > 0:
            batch_size = self.config.batch_size
        else:
            batch_size = DEFAULT_BATCH_SIZE
        while True:
            # Response format is {TopicPartiton('topic1', 1): [msg1, msg2]}
            msg_pack = self.consumer.poll(
                max_records=batch_size,
                timeout_ms=500,
            )

            message_values = []
            msg_printed = False
            for tp, messages in msg_pack.items():
                for message in messages:
                    if not msg_printed:
                        self.__print_message(message)
                        msg_printed = True
                    message_values.append(self.__deserialize_message(message.value))
            if len(message_values) > 0:
                handler(message_values)

    def test_connection(self):
        return True

    def __deserialize_message(self, message):
        if self.config.serde_config is not None and \
                self.config.serde_config.serialization_method == SerializationMethod.PROTOBUF and \
                self.schema_class is not None:
            from google.protobuf.json_format import MessageToDict
            obj = self.schema_class()
            obj.ParseFromString(message)
            return MessageToDict(obj)
        elif self.config.serde_config is not None and \
                self.config.serde_config.serialization_method == SerializationMethod.RAW_VALUE:
            return message
        else:
            return json.loads(message.decode('utf-8'))

    def __print_message(self, message):
        self._print(f'Receive message {message.partition}:{message.offset}: '
                    f'v={message.value}, time={time.time()}')



from mage_ai.api.operations import constants
from mage_ai.api.presenters.BasePresenter import BasePresenter


class IntegrationSourcePresenter(BasePresenter):
    default_attributes = [
        'name',
        'templates',
        'uuid',
    ]

    def present(self, **kwargs):
        if constants.UPDATE == kwargs['format']:
            payload = kwargs.get('payload', {})
            selected_streams = payload.get('integration_source', {}).get('streams')
            catalog = self.model.discover(streams=selected_streams) or {}

            return dict(
                selected_streams=selected_streams,
                streams=catalog.get('streams', []),
                uuid=self.model.source_uuid,
            )

        return self.model


IntegrationSourcePresenter.register_format(
    constants.CREATE,
    [
        'error_message',
        'streams',
        'success',
    ],
)

IntegrationSourcePresenter.register_format(
    constants.UPDATE,
    [
        'selected_streams',
        'streams',
        'uuid',
    ],
)



from mage_ai.api.errors import ApiError
from mage_ai.api.resources.GenericResource import GenericResource
from mage_ai.orchestration.db import safe_db_query
from mage_ai.data_preparation.models.widget import Widget
from mage_ai.data_preparation.repo_manager import get_repo_path
from typing import Dict
import asyncio


class WidgetResource(GenericResource):
    @classmethod
    @safe_db_query
    async def collection(self, query, meta, user, **kwargs):
        pipeline = kwargs['parent_model']

        include_outputs = query.get('include_outputs', [True])
        if include_outputs:
            include_outputs = include_outputs[0]

        collection = await asyncio.gather(
            *[widget.to_dict_async(
                include_content=True,
                include_outputs=include_outputs,
              ) for widget in pipeline.widgets_by_uuid.values()]
        )

        return self.build_result_set(
            collection,
            user,
            **kwargs,
        )

    @classmethod
    @safe_db_query
    def create(self, payload: Dict, user, **kwargs) -> 'WidgetResource':
        pipeline = kwargs['parent_model']

        resource = Widget.create(
            payload.get('name') or payload.get('uuid'),
            payload.get('type'),
            get_repo_path(),
            config=payload.get('config'),
            language=payload.get('language'),
            pipeline=pipeline,
            priority=payload.get('priority'),
            upstream_block_uuids=payload.get('upstream_blocks', []),
        )

        if payload.get('content'):
            resource.update_content(payload['content'], widget=True)

        if payload.get('configuration'):
            resource.configuration = payload['configuration']
            pipeline.save()

        return self(resource, user, **kwargs)

    @classmethod
    @safe_db_query
    async def member(self, pk, user, **kwargs):
        pipeline = kwargs['parent_model']
        widget = pipeline.get_block(pk, widget=True)

        error = ApiError.RESOURCE_NOT_FOUND.copy()
        if widget is None:
            error.update(message=f'Widget {pk} does not exist in pipeline {pipeline.uuid}.')
            raise ApiError(error)

        return self(widget, user, **kwargs)

    @safe_db_query
    def delete(self, **kwargs):
        self.model.delete()
        return self

    @safe_db_query
    def update(self, payload, **kwargs):
        self.model.update(payload)
        if payload.get('configuration'):
            self.model.configuration = payload['configuration']
            self.parent_model().save()
        return self



class Resource():
    def __init__(self, model, current_user, **kwargs):
        self.current_user = current_user
        self.model = model
        self.model_options = kwargs
        self.result_set_attr = None



from jupyter_server import subprocess
from logging import Logger
from mage_ai.data_integrations.logger.utils import (
    print_log_from_line,
)
from mage_ai.data_integrations.utils.config import (
    build_config_json,
    get_catalog_by_stream,
)
from mage_ai.data_preparation.models.block import PYTHON_COMMAND, Block
from mage_ai.data_preparation.models.constants import BlockType
from mage_ai.shared.hash import merge_dict
from typing import Dict, List

import json
import os
import pandas as pd


class IntegrationBlock(Block):
    def _execute_block(
        self,
        outputs_from_input_vars,
        execution_partition: str = None,
        input_vars: List = None,
        logger: Logger = None,
        logging_tags: Dict = dict(),
        global_vars: Dict = None,
        test_execution: bool = False,
        input_from_output: Dict = None,
        runtime_arguments: Dict = None,
        **kwargs,
    ) -> List:
        from mage_integrations.sources.constants import BATCH_FETCH_LIMIT

        index = self.template_runtime_configuration.get('index', None)
        is_last_block_run = self.template_runtime_configuration.get('is_last_block_run', False)
        selected_streams = self.template_runtime_configuration.get('selected_streams', [])
        stream = selected_streams[0] if len(selected_streams) >= 1 else None
        destination_table = self.template_runtime_configuration.get('destination_table', stream)
        query_data = runtime_arguments or {}
        query_data = query_data.copy()

        tags = dict(block_tags=dict(
            destination_table=destination_table,
            index=index,
            stream=stream,
            type=self.type,
            uuid=self.uuid,
        ))
        updated_logging_tags = merge_dict(
            logging_tags,
            dict(tags=tags),
        )

        if index is not None:
            source_state_file_path = self.pipeline.source_state_file_path(
                destination_table=destination_table,
                stream=stream,
            )
            destination_state_file_path = self.pipeline.destination_state_file_path(
                destination_table=destination_table,
                stream=stream,
            )
            source_output_file_path = self.pipeline.source_output_file_path(stream, index)

            stream_catalog = get_catalog_by_stream(
                self.pipeline.data_loader.file_path,
                stream,
                global_vars,
                pipeline=self.pipeline,
            ) or dict()

            if stream_catalog.get('replication_method') == 'INCREMENTAL':
                from mage_integrations.sources.utils import \
                    update_source_state_from_destination_state
                update_source_state_from_destination_state(
                    source_state_file_path,
                    destination_state_file_path,
                )
            else:
                query_data['_offset'] = BATCH_FETCH_LIMIT * index
            if not is_last_block_run:
                query_data['_limit'] = BATCH_FETCH_LIMIT

        outputs = []
        if BlockType.DATA_LOADER == self.type:
            lines_in_file = 0

            with open(source_output_file_path, 'w') as f:
                args = [
                    PYTHON_COMMAND,
                    self.pipeline.source_file_path,
                    '--config_json',
                    build_config_json(
                        self.pipeline.data_loader.file_path,
                        global_vars,
                    ),
                    '--log_to_stdout',
                    '1',
                    '--settings',
                    self.pipeline.settings_file_path,
                    '--state',
                    source_state_file_path,
                    '--query_json',
                    json.dumps(query_data),
                ]

                if len(selected_streams) >= 1:
                    args += [
                        '--selected_streams_json',
                        json.dumps(selected_streams),
                    ]

                proc = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

                for line in proc.stdout:
                    f.write(line.decode())
                    print_log_from_line(
                        line,
                        logger=logger,
                        logging_tags=logging_tags,
                        tags=tags,
                    )
                    lines_in_file += 1

                outputs.append(proc)

            proc.communicate()
            if proc.returncode != 0 and proc.returncode is not None:
                raise subprocess.CalledProcessError(proc.returncode, proc.args)

            file_size = os.path.getsize(source_output_file_path)
            msg = f'Finished writing {file_size} bytes with {lines_in_file} lines to output '\
                  f'file {source_output_file_path}.'
            if logger:
                logger.info(msg, **updated_logging_tags)
            else:
                print(msg)
        elif BlockType.TRANSFORMER == self.type:
            from mage_integrations.sources.constants import COLUMN_TYPE_NULL
            from mage_integrations.utils.logger.constants import (
                TYPE_RECORD,
                TYPE_SCHEMA,
            )
            from mage_integrations.transformers.utils import (
                convert_data_type,
                infer_dtypes,
            )

            decorated_functions = []
            test_functions = []

            results = {
                self.type: self._block_decorator(decorated_functions),
                'test': self._block_decorator(test_functions),
            }
            results.update(outputs_from_input_vars)

            exec(self.content, results)

            # 1. Recreate each record
            # 2. Recreate schema
            schema_original = None
            schema_updated = None
            schema_index = None
            output_arr = []
            records_transformed = 0
            df_sample = None

            with open(source_output_file_path, 'r') as f:
                idx = 0
                for line in f:
                    line = line.strip() if line else ''
                    if len(line) == 0:
                        continue

                    try:
                        data = json.loads(line)
                        line_type = data.get('type')

                        if TYPE_SCHEMA == line_type:
                            schema_index = idx
                            schema_original = data
                        elif TYPE_RECORD == line_type:
                            record = data['record']
                            input_vars = [pd.DataFrame.from_dict([record])]
                            input_kwargs = merge_dict(
                                global_vars,
                                dict(
                                    index=index,
                                    query=query_data,
                                    stream=stream,
                                ),
                            )
                            block_function = self._validate_execution(
                                decorated_functions,
                                input_vars,
                            )

                            if block_function is not None:
                                df = self.execute_block_function(
                                    block_function,
                                    input_vars,
                                    input_kwargs,
                                    test_execution,
                                )
                                if df_sample is None:
                                    df_sample = df

                                if not schema_updated:
                                    properties_updated = {
                                        k: dict(type=[COLUMN_TYPE_NULL, convert_data_type(v)])
                                        for k, v in infer_dtypes(df).items()
                                    }
                                    schema_updated = schema_original.copy()
                                    properties_original = schema_updated['schema']['properties']
                                    schema_updated['schema']['properties'] = {
                                        k: properties_original[k]
                                        if k in properties_original else v
                                        for k, v in properties_updated.items()
                                    }

                                if df.shape[0] == 0:
                                    continue
                                record_transformed = df.to_dict('records')[0]

                                line = json.dumps(merge_dict(
                                    data,
                                    dict(record=record_transformed),
                                ))
                                records_transformed += 1

                                if records_transformed % 1000 == 0:
                                    msg = f'{records_transformed} records have been transformed...'
                                    if logger:
                                        logger.info(msg, **updated_logging_tags)
                                    else:
                                        print(msg)
                    except json.decoder.JSONDecodeError:
                        pass

                    output_arr.append(line)
                    idx += 1

            output_arr[schema_index] = json.dumps(schema_updated)

            with open(source_output_file_path, 'w') as f:
                output = '\n'.join(output_arr)
                f.write(output)

            msg = f'Transformed {records_transformed} total records for stream {stream}.'
            file_size = os.path.getsize(source_output_file_path)
            msg2 = f'Finished writing {file_size} bytes with {len(output_arr)} lines to '\
                   f'output file {source_output_file_path}.'
            if logger:
                logger.info(msg, **updated_logging_tags)
                logger.info(msg2, **updated_logging_tags)
            else:
                print(msg)
                print(msg2)

            self.test_functions = test_functions
        elif BlockType.DATA_EXPORTER == self.type:
            override = {}
            if destination_table:
                override['table'] = destination_table

            file_size = os.path.getsize(source_output_file_path)
            msg = f'Reading {file_size} bytes from {source_output_file_path} as input file.'
            if logger:
                logger.info(msg, **updated_logging_tags)
            else:
                print(msg)

            proc = subprocess.Popen([
                PYTHON_COMMAND,
                self.pipeline.destination_file_path,
                '--config_json',
                build_config_json(
                    self.pipeline.data_exporter.file_path,
                    global_vars,
                    override=override,
                ),
                '--log_to_stdout',
                '1',
                '--settings',
                self.pipeline.data_exporter.file_path,
                '--state',
                self.pipeline.destination_state_file_path(
                    destination_table=destination_table,
                    stream=stream,
                ),
                '--input_file_path',
                source_output_file_path,
            ], stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

            for line in proc.stdout:
                print_log_from_line(
                    line,
                    logger=logger,
                    logging_tags=logging_tags,
                    tags=tags,
                )

            proc.communicate()
            if proc.returncode != 0 and proc.returncode is not None:
                raise subprocess.CalledProcessError(proc.returncode, proc.args)

            outputs.append(proc)

        return outputs


class SourceBlock(IntegrationBlock):
    pass


class DestinationBlock(IntegrationBlock):
    def to_dict(
        self,
        include_content=False,
        include_outputs=False,
        sample_count=None,
        check_if_file_exists: bool = False,
        destination_table: str = None,
        state_stream: str = None,
    ):
        data = {}
        if state_stream and destination_table:
            from mage_ai.data_preparation.models.pipelines.integration_pipeline \
                import IntegrationPipeline
            integration_pipeline = IntegrationPipeline(self.pipeline.uuid)
            destination_state_file_path = integration_pipeline.destination_state_file_path(
                destination_table=destination_table,
                stream=state_stream,
            )
            if os.path.isfile(destination_state_file_path):
                with open(destination_state_file_path, 'r') as f:
                    text = f.read()
                    d = json.loads(text) if text else {}
                    bookmark_values = d.get('bookmarks', {}).get(state_stream)
                    data['bookmarks'] = bookmark_values

        return merge_dict(
            super().to_dict(
                include_content,
                include_outputs,
                sample_count,
                check_if_file_exists,
            ),
            data,
        )

    def update(self, data, update_state=False):
        if update_state:
            from mage_ai.data_preparation.models.pipelines.integration_pipeline \
                import IntegrationPipeline
            from mage_integrations.destinations.utils \
                import update_destination_state_bookmarks

            integration_pipeline = IntegrationPipeline(self.pipeline.uuid)
            tap_stream_id = data.get('tap_stream_id')
            destination_table = data.get('destination_table')
            bookmark_values = data.get('bookmark_values', {})
            if tap_stream_id and destination_table:
                destination_state_file_path = integration_pipeline.destination_state_file_path(
                    destination_table=destination_table,
                    stream=tap_stream_id,
                )
                update_destination_state_bookmarks(
                    destination_state_file_path,
                    tap_stream_id,
                    bookmark_values=bookmark_values
                )

        return super().update(data)

    def output_variables(self, execution_partition: str = None) -> List[str]:
        return []


class TransformerBlock(IntegrationBlock):
    pass



from collections import UserList
from mage_ai import settings
from mage_ai.api.api_context import ApiContext
from mage_ai.api.errors import ApiError
from mage_ai.api.monitors.BaseMonitor import BaseMonitor
from mage_ai.api.operations.constants import (
    CREATE,
    DELETE,
    DETAIL,
    FILE_KEY_NAME,
    LIST,
    META_KEY_FORMAT,
    READ,
    UPDATE,
    WRITE,
)
from mage_ai.api.result_set import ResultSet
from mage_ai.orchestration.db.errors import DoesNotExistError
from mage_ai.shared.array import flatten
from mage_ai.shared.hash import merge_dict, ignore_keys
from typing import Dict
import dateutil.parser
import importlib
import inflection
import inspect


def classify(name):
    return ''.join([n.capitalize() for n in name.split('_')])


def singularize(name):
    return inflection.singularize(name)


class BaseOperation():
    def __init__(self, **kwargs):
        self.action = kwargs.get('action')
        self.files = kwargs.get('files', {})
        self.pk = kwargs.get('pk')
        self.headers = kwargs.get('headers', {})
        self.meta = kwargs.get('meta', {})
        self.oauth_client = kwargs.get('oauth_client')
        self.oauth_token = kwargs.get('oauth_token')
        self.options = kwargs.get('options', {})
        self.payload = kwargs.get('payload', {})
        self._query = kwargs.get('query', {}) or {}
        self.resource = kwargs.get('resource')
        self.resource_parent = kwargs.get('resource_parent')
        self.resource_parent_id = kwargs.get('resource_parent_id')
        self.user = kwargs.get('user')
        self.context = ApiContext()
        self.__combined_options_attr = None
        self.__presentation_format_attr = None
        self.__updated_options_attr = None

    async def execute(self):
        response = {}
        try:
            result = await self.__executed_result()
            presented = await self.__present_results(result)
            attrb = flatten([d.keys() for d in presented]) if issubclass(
                type(presented), list) else presented.keys()
            attrb = list(set(attrb))
            if (issubclass(type(result), list) or issubclass(type(result), UserList)):
                results = result
            else:
                results = [result]
            for res in results:
                updated_options = await self.__updated_options()
                policy = self.__policy_class()(res, self.user, **updated_options)
                policy.authorize_attributes(
                    READ, attrb, api_operation_action=self.action)
            response_key = self.resource if LIST == self.action else self.__resource_name_singular()
            response[response_key] = presented

            if isinstance(result, ResultSet):
                response['metadata'] = result.metadata

            if settings.DEBUG:
                debug_payload = {}
                debug_payload[self.__resource_name_singular()] = merge_dict(
                    self.payload.get(self.__resource_name_singular(), {}),
                    {
                        'file': None,
                    },
                )
                response['debug'] = {
                    'action': self.action,
                    'meta': self.meta,
                    'options': self.options,
                    'payload': merge_dict(self.payload, debug_payload),
                    'pk': self.pk,
                    'presentation_format': self.__presentation_format(),
                    'query': self.query,
                    'resource': self.resource,
                    'resource_parent': self.resource_parent,
                    'resource_parent_id': self.resource_parent_id,
                    'user': {
                        'id': self.user.id if self.user else None,
                    },
                }
        except ApiError as err:
            if settings.DEBUG:
                raise err
            else:
                response['error'] = self.__present_error(err)
        return response

    @property
    def query(self) -> Dict:
        if not self._query:
            return {}

        query = {}
        for key, values in self._query.items():
            query[key] = values
            if type(values) is list:
                arr = []
                for v in values:
                    try:
                        v = v.decode()
                    except (UnicodeDecodeError, AttributeError):
                        pass

                    if 'true' == v:
                        v = True
                    elif 'false' == v:
                        v = False

                    arr.append(v)
                query[key] = arr

        return query

    async def __executed_result(self):
        if self.action in [CREATE, LIST]:
            return await self.__create_or_index()
        elif self.action in [DELETE, DETAIL, UPDATE]:
            return await self.__delete_show_or_update()

    async def __create_or_index(self):
        updated_options = await self.__updated_options()
        policy = self.__policy_class()(None, self.user, **updated_options)
        policy.authorize_action(self.action)
        if CREATE == self.action:
            policy.authorize_attributes(
                WRITE,
                self.__payload_for_resource().keys(),
                **self.__payload_for_resource(),
            )
            options = updated_options.copy()
            options.pop('payload', None)
            return await self.__resource_class().process_create(
                self.__payload_for_resource(),
                self.user,
                **options,
            )
        elif LIST == self.action:
            policy.authorize_query(self.query)
            options = updated_options.copy()
            options.pop('meta', None)
            options.pop('query', None)
            return await self.__resource_class().process_collection(
                self.query,
                self.meta,
                self.user,
                **options,
            )

    async def __delete_show_or_update(self):
        updated_options = await self.__updated_options()
        res = await self.__resource_class().process_member(
            self.pk, self.user, **updated_options)

        policy = self.__policy_class()(res, self.user, **updated_options)
        policy.authorize_action(self.action)

        if DELETE == self.action:
            await res.process_delete(**updated_options)
        elif DETAIL == self.action:
            policy.authorize_query(self.query)
        elif UPDATE == self.action:
            policy.authorize_attributes(
                WRITE,
                self.__payload_for_resource().keys(),
                **self.__payload_for_resource(),
            )
            options = updated_options.copy()
            options.pop('payload', None)
            await res.process_update(self.__payload_for_resource(), **options)

        return res

    def __present_error(self, error):
        return self.__monitor_class()(
            self.resource,
            self.user,
            error,
            **self.__combined_options(),
        ).present()

    def __classified_class(self):
        return classify(self.__resource_name_singular())

    def __resource_name_singular(self):
        return singularize(self.resource)

    def __monitor_class(self):
        try:
            return getattr(
                importlib.import_module(
                    'mage_ai.api.monitors.{}Monitor'.format(
                        self.__classified_class())), '{}Monitor'.format(
                    self.__classified_class()), )
        except ModuleNotFoundError:
            return BaseMonitor

    def __policy_class(self):
        return getattr(
            importlib.import_module(
                'mage_ai.api.policies.{}Policy'.format(
                    self.__classified_class())), '{}Policy'.format(
                self.__classified_class()), )

    def __presenter_class(self):
        return getattr(
            importlib.import_module(
                'mage_ai.api.presenters.{}Presenter'.format(
                    self.__classified_class())), '{}Presenter'.format(
                self.__classified_class()), )

    def __resource_class(self):
        return getattr(
            importlib.import_module(
                'mage_ai.api.resources.{}Resource'.format(
                    self.__classified_class())), '{}Resource'.format(
                self.__classified_class()), )

    async def __parent_model(self):
        if self.resource_parent and self.resource_parent_id:
            parent_class = classify(singularize(self.resource_parent))
            parent_resource_class = getattr(
                importlib.import_module(
                    'mage_ai.api.resources.{}Resource'.format(parent_class)),
                '{}Resource'.format(parent_class),
            )
            try:
                model = parent_resource_class.get_model(self.resource_parent_id)
                if inspect.isawaitable(model):
                    model = await model
                return model
            except DoesNotExistError:
                raise ApiError(ApiError.RESOURCE_NOT_FOUND)

    def __combined_options(self):
        if not self.__combined_options_attr:
            self.__combined_options_attr = {
                'context': self.context,
                'meta': self.meta,
                'options': self.options,
                'payload': self.payload,
                'query': self.query,
            }
        return self.__combined_options_attr

    def __payload_for_resource(self):
        payload = self.payload.get(self.__resource_name_singular(), {})

        if self.files and self.files.get(FILE_KEY_NAME):
            payload_prev = ignore_keys(self.payload, [FILE_KEY_NAME])
            payload[FILE_KEY_NAME] = self.files.get(FILE_KEY_NAME)
            payload.update(payload_prev)

        for key in self.__resource_class().datetime_keys:
            if payload.get(key) is not None:
                if type(payload[key]) is str:
                    payload[key] = dateutil.parser.parse(payload[key])

        return payload

    async def __present_results(self, results):
        updated_options = await self.__updated_options()
        data = updated_options.copy()
        data.update({'format': self.__presentation_format()})
        return await self.__presenter_class().present_resource(results, self.user, **data)

    def __presentation_format(self):
        if not self.__presentation_format_attr:
            self.__presentation_format_attr = self.meta.get(
                META_KEY_FORMAT, self.action)
        return self.__presentation_format_attr

    async def __updated_options(self):
        if not self.__updated_options_attr:
            self.__updated_options_attr = self.__combined_options().copy()
            self.__updated_options_attr.update({
                'api_operation_action': self.action,
                'oauth_client': self.oauth_client,
                'oauth_token': self.oauth_token,
                'parent_model': await self.__parent_model(),
            })
            self.__updated_options_attr.update(self.headers)
        return self.__updated_options_attr



from mage_ai.api.oauth_scope import OauthScope
from mage_ai.api.operations import constants
from mage_ai.api.policies.BasePolicy import BasePolicy
from mage_ai.api.presenters.BlockPresenter import BlockPresenter


class BlockPolicy(BasePolicy):
    pass


BlockPolicy.allow_actions([
    constants.CREATE,
    constants.DELETE,
    constants.UPDATE,
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], condition=lambda policy: policy.has_at_least_editor_role_and_edit_access())

BlockPolicy.allow_actions([
    constants.DETAIL,
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], condition=lambda policy: policy.has_at_least_viewer_role())

BlockPolicy.allow_read([
    'content',
] + BlockPresenter.default_attributes, scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.CREATE,
    constants.UPDATE,
], condition=lambda policy: policy.has_at_least_editor_role_and_edit_access())

BlockPolicy.allow_read(BlockPresenter.default_attributes + [], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.DELETE,
], condition=lambda policy: policy.has_at_least_editor_role_and_edit_access())

BlockPolicy.allow_read([
    'bookmarks',
    'content',
    'outputs',
] + BlockPresenter.default_attributes, scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.DETAIL,
], condition=lambda policy: policy.has_at_least_viewer_role())

BlockPolicy.allow_write([
    'color',
    'config',
    'configuration',
    'content',
    'converted_from',
    'extension_uuid',
    'has_callback',
    'language',
    'metadata',
    'name',
    'priority',
    'type',
    'upstream_blocks',
    'uuid',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.CREATE,
], condition=lambda policy: policy.has_at_least_editor_role_and_edit_access())

BlockPolicy.allow_write([
    'all_upstream_blocks_executed',
    'bookmark_values',
    'color',
    'configuration',
    'content',
    'destination_table',
    'downstream_blocks',
    'executor_config',
    'executor_type',
    'extension_uuid',
    'has_callback',
    'language',
    'metadata',
    'name',
    'outputs',
    'status',
    'tags',
    'tap_stream_id',
    'type',
    'upstream_blocks',
    'uuid',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.UPDATE,
], condition=lambda policy: policy.has_at_least_editor_role_and_edit_access())

BlockPolicy.allow_query([
    'destination_table',
    'state_stream',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.DETAIL,
], condition=lambda policy: policy.has_at_least_viewer_role())

BlockPolicy.allow_query([
    'extension_uuid',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.DELETE,
    constants.UPDATE,
], condition=lambda policy: policy.has_at_least_viewer_role())



from mage_ai.api.presenters.BasePresenter import BasePresenter
from mage_ai.api.presenters.mixins.users import AssociatedUserPresenter


class SessionPresenter(BasePresenter, AssociatedUserPresenter):
    default_attributes = [
        'expires',
        'token',
        'user',
    ]



from collections.abc import Iterable
from mage_ai import settings
from mage_ai.api.errors import ApiError
from mage_ai.api.oauth_scope import OauthScope
from mage_ai.api.utils import (
    has_at_least_admin_role,
    has_at_least_editor_role,
    has_at_least_editor_role_and_edit_access,
    has_at_least_viewer_role,
    is_owner,
)
from mage_ai.services.tracking.metrics import increment
from mage_ai.settings import DISABLE_NOTEBOOK_EDIT_ACCESS
from mage_ai.shared.hash import extract
import importlib
import inflection

ALL_ACTIONS = 'all'


class BasePolicy():
    action_rules = {}
    query_rules = {}
    read_rules = {}
    write_rules = {}

    def __init__(self, resource, current_user, **kwargs):
        self.current_user = current_user
        self.options = kwargs
        self.resource = resource
        self.parent_model_attr = None
        self.parent_resource_attr = None
        if isinstance(resource, Iterable):
            self.resources = resource
        else:
            self.resources = [resource]

    @classmethod
    def action_rule(self, action):
        if not self.action_rules.get(self.__name__):
            self.action_rules[self.__name__] = {}
        return self.action_rules[self.__name__].get(action)

    @classmethod
    def query_rule(self, query):
        if not self.query_rules.get(self.__name__):
            self.query_rules[self.__name__] = {}
        return self.query_rules[self.__name__].get(query)

    @classmethod
    def read_rule(self, read):
        if not self.read_rules.get(self.__name__):
            self.read_rules[self.__name__] = {}
        return self.read_rules[self.__name__].get(read)

    @classmethod
    def write_rule(self, write):
        if not self.write_rules.get(self.__name__):
            self.write_rules[self.__name__] = {}
        return self.write_rules[self.__name__].get(write)

    @classmethod
    def allow_actions(self, array, **kwargs):
        if not self.action_rules.get(self.__name__):
            self.action_rules[self.__name__] = {}
        for key in array:
            if not self.action_rules[self.__name__].get(key):
                self.action_rules[self.__name__][key] = {}
            for scope in kwargs.get('scopes', []):
                self.action_rules[self.__name__][key][scope] = extract(kwargs, [
                                                                       'condition'])

    @classmethod
    def allow_query(self, array, **kwargs):
        if not self.query_rules.get(self.__name__):
            self.query_rules[self.__name__] = {}
        for key in array:
            if not self.query_rules[self.__name__].get(key):
                self.query_rules[self.__name__][key] = {}
            for scope in kwargs.get('scopes', []):
                self.query_rules[self.__name__][key][scope] = extract(kwargs, [
                                                                      'condition'])

    @classmethod
    def allow_read(self, array, **kwargs):
        if not self.read_rules.get(self.__name__):
            self.read_rules[self.__name__] = {}
        for key in array:
            if not self.read_rules[self.__name__].get(key):
                self.read_rules[self.__name__][key] = {}
            actions = kwargs.get('on_action', [ALL_ACTIONS])
            actions = actions if isinstance(actions, list) else [actions]
            for scope in kwargs.get('scopes', []):
                if not self.read_rules[self.__name__][key].get(scope):
                    self.read_rules[self.__name__][key][scope] = {}
                for action in actions:
                    self.read_rules[self.__name__][key][scope][action] = extract(kwargs, [
                                                                                 'condition'])

    @classmethod
    def allow_write(self, array, **kwargs):
        if not self.write_rules.get(self.__name__):
            self.write_rules[self.__name__] = {}
        for key in array:
            if not self.write_rules[self.__name__].get(key):
                self.write_rules[self.__name__][key] = {}
            actions = kwargs.get('on_action', [ALL_ACTIONS])
            actions = actions if isinstance(actions, list) else [actions]
            for scope in kwargs.get('scopes', []):
                if not self.write_rules[self.__name__][key].get(scope):
                    self.write_rules[self.__name__][key][scope] = {}
                for action in actions:
                    self.write_rules[self.__name__][key][scope][action] = extract(kwargs, [
                                                                                  'condition'])

    @classmethod
    def resource_name(self):
        return inflection.pluralize(self.resource_name_singular())

    @classmethod
    def resource_name_singular(self):
        return inflection.underscore(
            self.__name__.replace(
                'Policy', '')).lower()

    def is_owner(self) -> bool:
        return is_owner(self.current_user)

    def has_at_least_admin_role(self) -> bool:
        return has_at_least_admin_role(self.current_user)

    def has_at_least_editor_role(self) -> bool:
        return has_at_least_editor_role(self.current_user)

    def has_at_least_editor_role_and_edit_access(self) -> bool:
        return has_at_least_editor_role_and_edit_access(self.current_user)

    def has_at_least_viewer_role(self) -> bool:
        return has_at_least_viewer_role(self.current_user)

    def authorize_action(self, action):
        config = self.__class__.action_rule(action)
        if config:
            self.__validate_scopes(action, config.keys())
            if config.get(self.__current_scope(), {}).get('condition'):
                self.__validate_condition(
                    action, config[self.__current_scope()]['condition'])
        else:
            error = ApiError.UNAUTHORIZED_ACCESS
            error.update({'message': 'The {} action is disabled for {}.'.format(
                action, self.resource_name()), })
            raise ApiError(error)

    def authorize_attribute(self, read_or_write, attrb, **kwargs):
        if self.is_owner():
            return True

        api_operation_action = self.options.get(
            'api_operation_action',
            kwargs.get('api_operation_action', ALL_ACTIONS),
        )
        if 'read' == read_or_write:
            orig_config = self.__class__.read_rule(attrb)
        else:
            orig_config = self.__class__.write_rule(attrb)

        config = None
        if orig_config:
            self.__validate_scopes(attrb, orig_config.keys())
            config_scope = orig_config.get(self.__current_scope(), {})
            config = config_scope.get(api_operation_action)

            if config is None:
                config = config_scope.get(ALL_ACTIONS)

        if config is None:
            error = ApiError.UNAUTHORIZED_ACCESS
            error.update({
                'message': '{} of {} on {} is disabled for {}.'.format(
                    read_or_write,
                    attrb,
                    api_operation_action,
                    self.__class__.resource_name(),
                ),
            })
            raise ApiError(error)
        cond = config.get('condition')
        if cond:
            self.__validate_condition(attrb, cond, **kwargs)

    def authorize_attributes(self, read_or_write, attrbs, **kwargs):
        if self.is_owner():
            return True

        for attrb in attrbs:
            self.authorize_attribute(read_or_write, attrb, **kwargs)

    def authorize_query(self, query):
        if not query:
            return

        for key, value in query.items():
            if key != settings.QUERY_API_KEY:
                error_message = 'Query parameter {} of value {} is not permitted.'.format(
                    key, value)
                config = self.__class__.query_rule(key)

                if not config:
                    error = ApiError.UNAUTHORIZED_ACCESS
                    error.update({
                        'message': error_message,
                    })
                    raise ApiError(error)
                elif config.get(self.__current_scope(), {}).get('condition'):
                    self.__validate_condition(
                        key,
                        config[self.__current_scope()]['condition'],
                        message=error_message,
                    )

    def parent_model(self):
        if not self.parent_model_attr:
            self.parent_model_attr = self.options.get('parent_model', None)
        return self.parent_model_attr

    def parent_resource(self):
        if not self.parent_resource_attr and self.parent_model():
            parent_model_class = inflection.singularize(
                self.parent_model().__class__.__name__,
            )
            self.parent_resource_attr = getattr(
                importlib.import_module(
                    'mage_ai.api.resources.{}Resource'.format(parent_model_class)),
                '{}Resource'.format(parent_model_class),
            )(self.parent_model(), self.current_user, **self.options)
        return self.parent_resource_attr

    def __current_scope(self):
        if self.current_user or DISABLE_NOTEBOOK_EDIT_ACCESS:
            return OauthScope.CLIENT_PRIVATE
        else:
            return OauthScope.CLIENT_PUBLIC

    def __validate_condition(self, action, cond, **kwargs):
        if cond and not cond(self):
            error = ApiError.UNAUTHORIZED_ACCESS
            error.update({
                'message': kwargs.get(
                    'message',
                    'Unauthorized access for {}, failed condition.'.format(action),
                ),
            })
            increment(
                'api_error.unauthorized_access',
                tags={
                    'action': action,
                    'policy_class_name': self.__class__.__name__,
                    'id': self.resource.id if hasattr(
                        self.resource,
                        'id') else None,
                })
            if settings.DEBUG:
                error['debug'] = {
                    'action': action,
                }
            raise ApiError(error)

    def __validate_scopes(self, val, scopes):
        error = ApiError.UNAUTHORIZED_ACCESS
        if self.is_owner():
            return
        if OauthScope.CLIENT_ALL in scopes:
            return
        elif not self.current_user and OauthScope.CLIENT_PUBLIC not in scopes:
            error.update({
                'message': 'Private access for {} only, invalid scope.'.format(val),
            })
            if settings.DEBUG:
                error['debug'] = {
                    'class': self.__class__.__name__,
                    'scopes': scopes,
                }
            raise ApiError(error)
        elif self.current_user and OauthScope.CLIENT_PRIVATE not in scopes:
            error.update({
                'message': 'Public access for {} only, invalid scope.'.format(val),
            })
            if settings.DEBUG:
                error['debug'] = {
                    'scopes': scopes,
                }
            raise ApiError(error)



from mage_ai.api.errors import ApiError
from mage_ai.api.resources.GenericResource import GenericResource
from mage_ai.data_preparation.models.block import Block
from mage_ai.data_preparation.models.block.dbt import DBTBlock
from mage_ai.data_preparation.models.block.utils import clean_name
from mage_ai.data_preparation.models.constants import (
    BlockLanguage,
    BlockType,
    FILE_EXTENSION_TO_BLOCK_LANGUAGE,
)
from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.data_preparation.utils.block.convert_content import convert_to_block
from mage_ai.orchestration.db import safe_db_query
import urllib.parse


class BlockResource(GenericResource):
    @classmethod
    @safe_db_query
    def create(self, payload, user, **kwargs):
        pipeline = kwargs.get('parent_model')

        content = payload.get('content')
        name = payload.get('name')
        language = payload.get('language')
        uuid = clean_name(name)
        """
        New DBT models include "content" in its block create payload,
        whereas creating blocks from existing DBT model files do not.
        """
        if payload.get('type') == BlockType.DBT and content and language == BlockLanguage.SQL:
            dbt_block = DBTBlock(
                name,
                uuid,
                BlockType.DBT,
                configuration=payload.get('configuration'),
                language=language,
            )
            if dbt_block.file_path and dbt_block.file.exists():
                raise Exception('DBT model at that folder location already exists. \
                    Please choose a different model name, or add a DBT model by \
                    selecting single model from file.')

        block = Block.create(
            name or payload.get('uuid'),
            payload.get('type'),
            get_repo_path(),
            color=payload.get('color'),
            config=payload.get('config'),
            configuration=payload.get('configuration'),
            extension_uuid=payload.get('extension_uuid'),
            language=language,
            pipeline=pipeline,
            priority=payload.get('priority'),
            upstream_block_uuids=payload.get('upstream_blocks', []),
        )

        if content:
            if payload.get('converted_from'):
                content = convert_to_block(block, content)

            block.update_content(content)

        return self(block, user, **kwargs)

    @classmethod
    @safe_db_query
    def member(self, pk, user, **kwargs):
        error = ApiError.RESOURCE_INVALID.copy()

        query = kwargs.get('query', {})
        extension_uuid = query.get('extension_uuid', [None])
        if extension_uuid:
            extension_uuid = extension_uuid[0]

        pipeline = kwargs.get('parent_model')
        if pipeline:
            block = pipeline.get_block(pk, extension_uuid=extension_uuid)
            if block:
                return self(block, user, **kwargs)
            else:
                if extension_uuid:
                    message = f'Block {pk} does not exist in pipeline {pipeline.uuid} ' \
                        f'for extension {extension_uuid}.'
                else:
                    message = f'Block {pk} does not exist in pipeline {pipeline.uuid}.'
                error.update(message=message)
                raise ApiError(error)

        block_type_and_uuid = urllib.parse.unquote(pk)
        parts = block_type_and_uuid.split('/')

        if len(parts) < 2:
            error.update(message='The url path should be in block_type/block_uuid format.')
            raise ApiError(error)

        block_type = parts[0]
        block_uuid_with_extension = '/'.join(parts[1:])
        parts2 = block_uuid_with_extension.split('.')
        language = None
        if len(parts2) >= 2:
            block_uuid = '.'.join(parts2[:-1])
            language = FILE_EXTENSION_TO_BLOCK_LANGUAGE[parts2[-1]]
        else:
            block_uuid = block_uuid_with_extension

        if BlockType.DBT == block_type:
            block = DBTBlock(
                block_uuid,
                block_uuid,
                block_type,
                configuration=dict(file_path=block_uuid_with_extension),
                language=language,
            )
        else:
            block = Block.get_block(block_uuid, block_uuid, block_type, language=language)

        if not block.exists():
            error.update(ApiError.RESOURCE_NOT_FOUND)
            raise ApiError(error)

        return self(block, user, **kwargs)

    @safe_db_query
    def delete(self, **kwargs):
        query = kwargs.get('query', {})

        force = query.get('force', [False])
        if force:
            force = force[0]
        return self.model.delete(force=force)

    @safe_db_query
    def update(self, payload, **kwargs):
        query = kwargs.get('query', {})
        update_state = query.get('update_state', [False])
        if update_state:
            update_state = update_state[0]
        self.model.update(
            payload,
            update_state=update_state,
        )



from mage_ai.shared.constants import (
    ENV_DEV,
    ENV_PROD,
    ENV_STAGING,
    ENV_TEST,
)
import os
import sys


def is_debug():
    return int(os.getenv('DEBUG', 0)) == 1


def is_dev():
    return os.getenv('ENV', None) == 'dev'


def is_test():
    return os.getenv('ENV', None) == 'test' or any('unittest' in v for v in sys.argv)


def is_production():
    return os.getenv('ENV', None) == 'production'


def is_staging():
    return os.getenv('ENV', None) == 'staging'


def get_env():
    if is_test():
        return ENV_TEST
    elif is_production():
        return ENV_PROD
    elif is_staging():
        return ENV_STAGING
    else:
        return ENV_DEV



{% extends "data_loaders/default.jinja" %}
{% block imports %}
from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.io.config import ConfigFileLoader
from mage_ai.io.azure_blob_storage import AzureBlobStorage
from os import path
{{ super() -}}
{% endblock %}


{% block content %}
@data_loader
def load_from_azure_blob_storage(*args, **kwargs):
    """
    Template for loading data from a Azure Blob Storage.
    Specify your configuration settings in 'io_config.yaml'.

    Docs: https://docs.mage.ai/design/data-loading#azureblobstorage
    """
    config_path = path.join(get_repo_path(), 'io_config.yaml')
    config_profile = 'default'

    container_name = 'your_container_name'
    blob_path = 'your_blob_path'

    return AzureBlobStorage.with_config(ConfigFileLoader(config_path, config_profile)).load(
        container_name,
        blob_path,
    )
{% endblock %}



from mage_ai.data_preparation.models.constants import BlockType
from mage_ai.data_preparation.variable_manager import get_variable
from mage_ai.data_preparation.models.block.sql.utils.shared import (
    blocks_in_query,
    interpolate_input,
    should_cache_data_from_upstream,
    table_name_parts,
)
from mage_ai.io.config import ConfigKey
from pandas import DataFrame
from typing import Dict


def create_upstream_block_tables(
    loader,
    block,
    configuration: Dict = None,
    execution_partition: str = None,
    cache_upstream_dbt_models: bool = False,
    query: str = None,
):
    from mage_ai.data_preparation.models.block.dbt.utils import (
        parse_attributes,
        source_table_name_for_block,
    )
    configuration = configuration if configuration else block.configuration
    database_default = configuration.get('data_provider_database') or loader.default_database()

    mapping = blocks_in_query(block, query)
    for idx, upstream_block in enumerate(block.upstream_blocks):
        if query and upstream_block.uuid not in mapping:
            continue

        if should_cache_data_from_upstream(block, upstream_block, [
            'data_provider',
        ], [
            ConfigKey.GOOGLE_SERVICE_ACC_KEY,
            ConfigKey.GOOGLE_SERVICE_ACC_KEY_FILEPATH,
        ]):
            if BlockType.DBT == upstream_block.type and not cache_upstream_dbt_models:
                continue

            df = get_variable(
                upstream_block.pipeline.uuid,
                upstream_block.uuid,
                'output_0',
                partition=execution_partition,
            )

            no_data = False
            if type(df) is DataFrame:
                if len(df.index) == 0:
                    no_data = True
            elif type(df) is dict and len(df) == 0:
                no_data = True
            elif type(df) is list and len(df) == 0:
                no_data = True
            elif not df:
                no_data = True

            if no_data:
                print(f'\n\nNo data in upstream block {upstream_block.uuid}.')
                continue

            database_custom, schema_name, table_name = table_name_parts(
                configuration,
                upstream_block,
            )

            database = database_custom or database_default

            if BlockType.DBT == block.type and BlockType.DBT != upstream_block.type:
                attributes_dict = parse_attributes(block)
                schema_name = attributes_dict['source_name']
                table_name = source_table_name_for_block(upstream_block)

            full_table_name = f'{schema_name}.{table_name}'
            print(f'\n\nExporting data from upstream block {upstream_block.uuid} '
                  f'to {database}.{full_table_name}.')

            loader.export(
                df,
                full_table_name,
                database=database,
                if_exists='replace',
                verbose=False,
            )


def interpolate_input_data(block, query, loader):
    return interpolate_input(
        block,
        query,
        get_database=lambda opts: loader.default_database(),
    )



from kubernetes import client, config
from mage_ai.services.k8s.config import K8sExecutorConfig
from mage_ai.services.k8s.constants import (
    DEFAULT_NAMESPACE,
    KUBE_POD_NAME_ENV_VAR,
)
from mage_ai.shared.hash import merge_dict
import os
import time


class JobManager():
    def __init__(
        self,
        job_name: str = 'mage-job',
        namespace: str = DEFAULT_NAMESPACE,
        logger=None,
        logging_tags=dict(),
    ):
        self.job_name = job_name
        self.namespace = namespace
        self.logger = logger
        self.logging_tags = logging_tags or dict()

        self.load_config()
        self.batch_api_client = client.BatchV1Api()
        self.api_version = 'batch/v1'
        self.core_api_client = client.CoreV1Api()

        self.pod_config = self.core_api_client.read_namespaced_pod(
            name=os.getenv(KUBE_POD_NAME_ENV_VAR),
            namespace=self.namespace,
        )

    @classmethod
    def load_config(cls) -> bool:
        try:
            config.load_incluster_config()
            return True
        except Exception:
            pass

        try:
            config.load_kube_config()
            return True
        except Exception:
            pass

        return False

    def run_job(
        self,
        command: str,
        k8s_config=None,
    ):
        if type(k8s_config) is dict:
            k8s_config = K8sExecutorConfig.load(config=k8s_config)
        job = self.create_job_object(
            command,
            k8s_config=k8s_config,
        )

        self.create_job(job)

        api_response = None
        job_completed = False
        while not job_completed:
            api_response = self.batch_api_client.read_namespaced_job_status(
                name=self.job_name,
                namespace=self.namespace
            )
            if api_response.status.succeeded is not None or \
                    api_response.status.failed is not None:
                job_completed = True
            time.sleep(5)
            # self._print(f'Job {self.job_name} status={api_response.status}')

        self.delete_job()
        self._print(f'Job {self.job_name} status={api_response.status}')
        if api_response.status.succeeded is None:
            raise Exception(f'Failed to execute k8s job {self.job_name}')

    def create_job_object(
        self,
        command: str,
        k8s_config: K8sExecutorConfig = None,
    ):
        # Configureate Pod template container
        mage_server_container_spec = self.pod_config.spec.containers[0]

        container_kwargs = dict(
            name='mage-job-container',
            image=mage_server_container_spec.image,
            image_pull_policy='IfNotPresent',
            command=command.split(' ') if isinstance(command, str) else command,
            env=mage_server_container_spec.env,
            volume_mounts=mage_server_container_spec.volume_mounts,
        )
        if k8s_config and (k8s_config.resource_limits or k8s_config.resource_requests):
            resource_kwargs = dict()
            if k8s_config.resource_limits:
                resource_kwargs['limits'] = k8s_config.resource_limits.to_dict()
            if k8s_config.resource_requests:
                resource_kwargs['requests'] = k8s_config.resource_requests.to_dict()
            container_kwargs['resources'] = client.V1ResourceRequirements(
                **resource_kwargs,
            )

        container = client.V1Container(
            **container_kwargs,
        )
        # Create and configurate a spec section
        template = client.V1PodTemplateSpec(
            metadata=client.V1ObjectMeta(labels={'name': self.job_name}),
            spec=client.V1PodSpec(
                restart_policy='Never',
                containers=[container],
                volumes=self.pod_config.spec.volumes,
            ),
        )
        # Create the specification of deployment
        spec = client.V1JobSpec(template=template, backoff_limit=0)
        # Instantiate the job object
        job = client.V1Job(
            api_version=self.api_version,
            kind='Job',
            metadata=client.V1ObjectMeta(name=self.job_name),
            spec=spec)

        return job

    def create_job(self, job):
        api_response = self.batch_api_client.create_namespaced_job(
            body=job,
            namespace=self.namespace,
        )
        self._print("Job created. status='%s'" % str(api_response.status))

    def delete_job(self):
        api_response = self.batch_api_client.delete_namespaced_job(
            name=self.job_name,
            namespace=self.namespace,
            body=client.V1DeleteOptions(
                propagation_policy='Foreground',
                grace_period_seconds=0))
        self._print("Job deleted. status='%s'" % str(api_response.status))

    def _print(self, message, **kwargs):
        if self.logger is None:
            print(message, **kwargs)
        else:
            self.logger.info(message, **merge_dict(self.logging_tags, kwargs))



from jinja2 import Template
from mage_ai.data_preparation.shared.constants import REPO_PATH_ENV_VAR
from mage_ai.data_preparation.templates.utils import copy_template_directory
from mage_ai.shared.environments import is_test
from typing import Dict
import os
import sys
import traceback
import yaml

MAGE_DATA_DIR_ENV_VAR = 'MAGE_DATA_DIR'
if is_test():
    DEFAULT_MAGE_DATA_DIR = './'
else:
    DEFAULT_MAGE_DATA_DIR = '~/.mage_data'

DEFAULT_VARIABLE_RETENTION_PERIOD = '30d'


class RepoConfig:
    def __init__(self, repo_path: str = None, config_dict: Dict = None):
        from mage_ai.data_preparation.shared.utils import get_template_vars
        self.repo_path = repo_path or get_repo_path()
        self.repo_name = os.path.basename(self.repo_path)
        try:
            if not config_dict:
                metadata_path = os.path.join(self.repo_path, 'metadata.yaml')
                if os.path.exists(metadata_path):
                    with open(os.path.join(self.repo_path, 'metadata.yaml')) as f:
                        config_file = Template(f.read()).render(
                            **get_template_vars()
                        )
                        repo_config = yaml.full_load(config_file) or {}
                else:
                    repo_config = dict()
            else:
                repo_config = config_dict

            # Priority:
            # 1. 'variables_dir' from config_dict
            # 1. os.getenv(MAGE_DATA_DIR_ENV_VAR)
            # 2. 'variables_dir' from project's metadata.yaml file
            # 3. DEFAULT_MAGE_DATA_DIR
            if config_dict is not None and config_dict.get('variables_dir'):
                self.variables_dir = config_dict.get('variables_dir')
            elif os.getenv(MAGE_DATA_DIR_ENV_VAR):
                self.variables_dir = os.getenv(MAGE_DATA_DIR_ENV_VAR)
            else:
                self.variables_dir = os.path.expanduser(
                    repo_config.get('variables_dir', DEFAULT_MAGE_DATA_DIR),
                )
            if self.variables_dir is not None and not self.variables_dir.startswith('s3'):
                if os.path.isabs(self.variables_dir) and self.variables_dir != self.repo_path and (
                    not config_dict or not config_dict.get('variables_dir')
                ):
                    # If the variables_dir is an absolute path, not same as repo_path, and
                    # from config file
                    self.variables_dir = os.path.join(self.variables_dir, self.repo_name)
                else:
                    self.variables_dir = os.path.abspath(
                        os.path.join(self.repo_path, self.variables_dir),
                    )
            os.makedirs(self.variables_dir, exist_ok=True)

            self.remote_variables_dir = repo_config.get('remote_variables_dir')
            self.ecs_config = repo_config.get('ecs_config')
            self.emr_config = repo_config.get('emr_config')
            self.gcp_cloud_run_config = repo_config.get('gcp_cloud_run_config')
            self.k8s_executor_config = repo_config.get('k8s_executor_config')
            self.notification_config = repo_config.get('notification_config', dict())
            self.queue_config = repo_config.get('queue_config', dict())

            self.s3_bucket = None
            self.s3_path_prefix = None
            if self.remote_variables_dir is not None and \
                    self.remote_variables_dir.startswith('s3://'):
                path_parts = self.remote_variables_dir.replace('s3://', '').split('/')
                self.s3_bucket = path_parts.pop(0)
                self.s3_path_prefix = '/'.join(path_parts)

            self.logging_config = repo_config.get('logging_config', dict())

            self.variables_retention_period = repo_config.get(
                'variables_retention_period',
                DEFAULT_VARIABLE_RETENTION_PERIOD,
            )
        except Exception:
            traceback.print_exc()
            pass

    @classmethod
    def from_dict(self, config_dict: Dict) -> 'RepoConfig':
        repo_path = config_dict.get('repo_path')
        repo_config = RepoConfig(repo_path=repo_path, config_dict=config_dict)
        return repo_config

    def to_dict(self, remote: bool = False) -> Dict:
        return dict(
            ecs_config=self.ecs_config,
            emr_config=self.emr_config,
            gcp_cloud_run_config=self.gcp_cloud_run_config,
            notification_config=self.notification_config,
            queue_config=self.queue_config,
            repo_path=self.repo_path,
            variables_dir=self.remote_variables_dir if remote else self.variables_dir,
            variables_retention_period=self.variables_retention_period,
            remote_variables_dir=self.remote_variables_dir,
        )


def init_repo(repo_path: str) -> None:
    """
    Initialize a repository under the current path.
    """
    if os.path.exists(repo_path):
        raise FileExistsError(f'Repository {repo_path} already exists')

    os.makedirs(
        os.getenv(MAGE_DATA_DIR_ENV_VAR) or DEFAULT_MAGE_DATA_DIR,
        exist_ok=True,
    )
    copy_template_directory('repo', repo_path)


def get_data_dir() -> str:
    return os.getenv(MAGE_DATA_DIR_ENV_VAR) or DEFAULT_MAGE_DATA_DIR


def get_repo_name() -> str:
    return os.path.basename(get_repo_path())


def get_repo_path() -> str:
    return os.getenv(REPO_PATH_ENV_VAR) or os.getcwd()


def get_repo_config(repo_path=None) -> RepoConfig:
    return RepoConfig(repo_path=repo_path)


def set_repo_path(repo_path: str) -> None:
    os.environ[REPO_PATH_ENV_VAR] = repo_path
    sys.path.append(os.path.dirname(repo_path))


def get_variables_dir(repo_path: str = None) -> str:
    return get_repo_config(repo_path=repo_path).variables_dir



from mage_ai.streaming.sources.azure_event_hub import AzureEventHubSource
from mage_ai.tests.base_test import TestCase
from unittest.mock import patch


class KafkaTests(TestCase):
    def test_init(self):
        with patch.object(AzureEventHubSource, 'init_client') as mock_init_client:
            AzureEventHubSource(dict(
                connection_str='test_connection_str',
                eventhub_name='test_eventhub_name',
            ))
            mock_init_client.assert_called_once()

    def test_init_invalid_config(self):
        with patch.object(AzureEventHubSource, 'init_client') as mock_init_client:
            with self.assertRaises(Exception) as context:
                AzureEventHubSource(dict(
                    connection_str='test_connection_str',
                ))
            self.assertTrue(
                '__init__() missing 1 required positional argument: \'eventhub_name\''
                in str(context.exception),
            )
            self.assertEqual(mock_init_client.call_count, 0)



class ConnectionDeprecated(Exception):
    pass


class ConnectionInactive(Exception):
    pass


class ConnectionNotFound(Exception):
    pass


class JobNotFound(Exception):
    pass


class SyncJobFailed(Exception):
    pass


class UnhealthyServer(Exception):
    pass



from mage_ai.data_preparation.executors.block_executor import BlockExecutor
from mage_ai.data_preparation.executors.pipeline_executor import PipelineExecutor
from mage_ai.data_preparation.models.constants import (
    BlockType,
    ExecutorType,
    PipelineType,
)
from mage_ai.data_preparation.models.pipeline import Pipeline
from mage_ai.shared.code import is_pyspark_code
from typing import Union
import os


class ExecutorFactory:
    @classmethod
    def get_default_executor_type(self):
        executor_type = os.getenv('DEFAULT_EXECUTOR_TYPE', ExecutorType.LOCAL_PYTHON)
        if ExecutorType.is_valid_type(executor_type):
            return executor_type
        return ExecutorType.LOCAL_PYTHON

    @classmethod
    def get_pipeline_executor(
        self,
        pipeline: Pipeline,
        execution_partition: Union[str, None] = None,
    ) -> PipelineExecutor:
        if pipeline.type == PipelineType.PYSPARK:
            from mage_ai.data_preparation.executors.pyspark_pipeline_executor import (
                PySparkPipelineExecutor,
            )
            return PySparkPipelineExecutor(pipeline)
        elif pipeline.type == PipelineType.STREAMING:
            from mage_ai.data_preparation.executors.streaming_pipeline_executor import (
                StreamingPipelineExecutor,
            )
            return StreamingPipelineExecutor(pipeline, execution_partition=execution_partition)
        else:
            return PipelineExecutor(pipeline)

    @classmethod
    def get_block_executor(
        self,
        pipeline: Pipeline,
        block_uuid: str,
        execution_partition: Union[str, None] = None,
        executor_type: Union[ExecutorType, str, None] = None,
    ) -> BlockExecutor:
        executor_kwargs = dict(
            pipeline=pipeline,
            block_uuid=block_uuid,
            execution_partition=execution_partition,
        )
        if executor_type is None:
            block = pipeline.get_block(block_uuid, check_template=True)
            if pipeline.type == PipelineType.PYSPARK and (
                block.type != BlockType.SENSOR or is_pyspark_code(block.content)
            ):
                executor_type = ExecutorType.PYSPARK
            else:
                executor_type = block.executor_type
                if executor_type == ExecutorType.LOCAL_PYTHON:
                    # Use default executor type
                    executor_type = self.get_default_executor_type()
        if executor_type == ExecutorType.PYSPARK:
            from mage_ai.data_preparation.executors.pyspark_block_executor import (
                PySparkBlockExecutor,
            )
            return PySparkBlockExecutor(**executor_kwargs)
        elif executor_type == ExecutorType.ECS:
            from mage_ai.data_preparation.executors.ecs_block_executor import (
                EcsBlockExecutor,
            )
            return EcsBlockExecutor(**executor_kwargs)
        elif executor_type == ExecutorType.GCP_CLOUD_RUN:
            from mage_ai.data_preparation.executors.gcp_cloud_run_block_executor import (
                GcpCloudRunBlockExecutor,
            )
            return GcpCloudRunBlockExecutor(**executor_kwargs)
        elif executor_type == ExecutorType.K8S:
            from mage_ai.data_preparation.executors.k8s_block_executor import (
                K8sBlockExecutor,
            )
            return K8sBlockExecutor(**executor_kwargs)
        else:
            return BlockExecutor(**executor_kwargs)



{% extends "data_loaders/pyspark/default.jinja" %}
{% block imports %}
{{ super() -}}
{% endblock %}


{% block content %}
@data_loader
def load_from_s3_bucket(**kwargs):
    """
    Template for loading data from a S3 bucket.
    """
    s3_bucket = 'your_s3_bucket_name'
    s3_path_prefix = 'your_s3_path_prefix'
    df = (kwargs['spark'].read
        .format('csv')
        .option('header', 'true')
        .option('inferSchema', 'true')
        .option('delimiter', ',')
        .load(f's3://{s3_bucket}/{s3_path_prefix}/*')
    )

    return df
{% endblock %}



from dataclasses import dataclass
from enum import Enum


class SerializationMethod(str, Enum):
    JSON = 'JSON'
    PROTOBUF = 'PROTOBUF'
    RAW_VALUE = 'RAW_VALUE'


@dataclass
class SerDeConfig:
    serialization_method: SerializationMethod
    schema_classpath: str = None



from mage_ai.data_cleaner.cleaning_rules.remove_duplicate_rows import RemoveDuplicateRows
from mage_ai.tests.base_test import TestCase
import pandas as pd


class RemoveDuplicateRowsTests(TestCase):
    def test_evaluate(self):
        df = pd.DataFrame(
            [
                [1, '2022-01-01'],
                [2, '2022-01-02'],
                [3, '2022-01-03'],
                [2, '2022-01-02'],
                [4, '2022-01-04'],
                [5, '2022-01-05'],
                [3, '2022-01-03'],
            ],
            columns=['id', 'deleted_at'],
        )
        column_types = {
            'id': 'number',
            'deleted_at': 'datetime',
        }
        result = RemoveDuplicateRows(
            df,
            column_types,
            {},
        ).evaluate()

        self.assertEqual(
            result,
            [
                dict(
                    title='Remove duplicate rows',
                    message='Remove 2 duplicated row(s) to reduce the amount of redundant data.',
                    status='not_applied',
                    action_payload=dict(
                        action_type='drop_duplicate',
                        action_arguments=[],
                        action_code='',
                        action_options={},
                        action_variables={},
                        axis='row',
                        outputs=[],
                    ),
                )
            ],
        )



"""Add upstream failed to block run status

Revision ID: 847ea6602d6f
Revises: 1f9353eddbc6
Create Date: 2023-03-01 11:17:44.684404

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '847ea6602d6f'
down_revision = '1f9353eddbc6'
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    bind = op.get_bind()
    if bind.engine.name == 'postgresql':
        with op.get_context().autocommit_block():
            op.execute("ALTER TYPE blockrunstatus ADD VALUE 'UPSTREAM_FAILED'")
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    bind = op.get_bind()
    if bind.engine.name == 'postgresql':
        op.execute("ALTER TYPE blockrunstatus RENAME TO blockrunstatus_old")
        op.execute("CREATE TYPE blockrunstatus AS ENUM('INITIAL', 'QUEUED', 'RUNNING', 'COMPLETED', 'FAILED', 'CANCELLED')")
        op.execute((
            "ALTER TABLE block_run ALTER COLUMN status TYPE blockrunstatus USING "
            "status::text::blockrunstatus"
        ))
        op.execute("DROP TYPE blockrunstatus_old")
    # ### end Alembic commands ###



from mage_ai.api.utils import authenticate_client_and_token
from mage_ai.orchestration.db.models.oauth import Oauth2Application
from mage_ai.settings import REQUIRE_USER_AUTHENTICATION
import terminado
import tornado.websocket
import re


class TerminalWebsocketServer(terminado.TermSocket):
    @property
    def term_command(self):
        return next(iter(self.term_manager.shell_command))

    def check_origin(self, origin):
        return True

    def on_pty_read(self, text):
        """Data read from pty; send to frontend"""
        updated_text = text
        if self.term_command == 'cmd':
            xterm_escape = re.compile(r'(?:\x1B\]0;).*\x07')
            updated_text = xterm_escape.sub('', text)
        self.send_json_message(["stdout", updated_text])

    def open(self, *args, **kwargs):
        """Websocket connection opened.

        Call our terminal manager to get a terminal, and connect to it as a
        client.
        """
        # Jupyter has a mixin to ping websockets and keep connections through
        # proxies alive. Call super() to allow that to set up:
        tornado.websocket.WebSocketHandler.open(self, *args, **kwargs)

        api_key = self.get_argument('api_key', None, True)
        token = self.get_argument('token', None, True)

        user = None
        if REQUIRE_USER_AUTHENTICATION and api_key and token:
            oauth_client = Oauth2Application.query.filter(
                Oauth2Application.client_id == api_key,
            ).first()
            if oauth_client:
                oauth_token, valid = authenticate_client_and_token(oauth_client.id, token)
                valid = valid and \
                    oauth_token and \
                    oauth_token.user
                if valid:
                    user = oauth_token.user

        self.term_name = "tty"
        if user:
            self.term_name = f'terminal_{user.id}'

        self._logger.info("TermSocket.open: %s", self.term_name)

        self.terminal = self.term_manager.get_terminal(self.term_name)
        self.terminal.clients.append(self)
        self.__initiate_terminal(self.terminal)

    def __initiate_terminal(self, terminal):
        self.send_json_message(["setup", {}])
        self._logger.info("TermSocket.open: Opened %s", self.term_name)
        # Now drain the preopen buffer, if reconnect.
        buffered = ""
        preopen_buffer = terminal.read_buffer.copy()
        while True:
            if not preopen_buffer:
                break
            s = preopen_buffer.popleft()
            buffered += s
        if buffered:
            self.on_pty_read(buffered)

        # Turn enable-bracketed-paste off since it can mess up the output.
        if self.term_command == 'bash':
            terminal.ptyproc.write(
                "bind 'set enable-bracketed-paste off' # Mage terminal settings command\r")
        terminal.read_buffer.clear()



from croniter import croniter
from datetime import datetime, timedelta, timezone
from mage_ai.data_preparation.logging.logger_manager_factory import LoggerManagerFactory
from mage_ai.data_preparation.models.block.utils import (
    get_all_ancestors,
    is_dynamic_block,
)
from mage_ai.data_preparation.models.pipeline import Pipeline
from mage_ai.data_preparation.models.triggers import (
    ScheduleInterval,
    ScheduleStatus,
    ScheduleType,
    SettingsConfig,
    Trigger,
)
from mage_ai.orchestration.db import db_connection, safe_db_query
from mage_ai.orchestration.db.models.base import Base, BaseModel
from mage_ai.shared.array import find
from mage_ai.shared.dates import compare
from mage_ai.shared.hash import ignore_keys, index_by
from mage_ai.shared.utils import clean_name
from sqlalchemy import (
    Column,
    Boolean,
    DateTime,
    Enum,
    ForeignKey,
    Integer,
    JSON,
    String,
    Table,
)
from sqlalchemy.orm import joinedload, relationship, validates
from sqlalchemy.sql import func
from typing import Dict, List
import enum
import pytz
import traceback
import uuid


pipeline_schedule_event_matcher_association_table = Table(
    'pipeline_schedule_event_matcher_association',
    Base.metadata,
    Column('pipeline_schedule_id', ForeignKey('pipeline_schedule.id')),
    Column('event_matcher_id', ForeignKey('event_matcher.id')),
)


class PipelineSchedule(BaseModel):
    name = Column(String(255))
    pipeline_uuid = Column(String(255))
    schedule_type = Column(Enum(ScheduleType))
    start_time = Column(DateTime(timezone=True), default=None)
    schedule_interval = Column(String(50))
    status = Column(Enum(ScheduleStatus), default=ScheduleStatus.INACTIVE)
    variables = Column(JSON)
    sla = Column(Integer, default=None)  # in seconds
    token = Column(String(255), index=True, default=None)
    settings = Column(JSON)

    backfills = relationship('Backfill', back_populates='pipeline_schedule')
    pipeline_runs = relationship('PipelineRun', back_populates='pipeline_schedule')

    event_matchers = relationship(
        'EventMatcher',
        secondary=pipeline_schedule_event_matcher_association_table,
        back_populates='pipeline_schedules'
    )

    def get_settings(self) -> 'SettingsConfig':
        settings = self.settings if self.settings else dict()
        return SettingsConfig.load(config=settings)

    @property
    def pipeline(self) -> 'Pipeline':
        return Pipeline.get(self.pipeline_uuid)

    @property
    def pipeline_runs_count(self) -> int:
        return len(self.pipeline_runs)

    @validates('schedule_interval')
    def validate_schedule_interval(self, key, schedule_interval):
        if schedule_interval and schedule_interval not in \
                [e.value for e in ScheduleInterval]:
            if not croniter.is_valid(schedule_interval):
                raise ValueError('Cron expression is invalid.')

        return schedule_interval

    @property
    def last_pipeline_run_status(self) -> str:
        if len(self.pipeline_runs) == 0:
            return None
        return sorted(self.pipeline_runs, key=lambda x: x.created_at)[-1].status

    @classmethod
    @safe_db_query
    def active_schedules(self, pipeline_uuids: List[str] = None) -> List['PipelineSchedule']:
        query = self.query.filter(self.status == ScheduleStatus.ACTIVE)
        if pipeline_uuids is not None:
            query = query.filter(PipelineSchedule.pipeline_uuid.in_(pipeline_uuids))
        return query.all()

    @classmethod
    def create(self, **kwargs) -> 'PipelineSchedule':
        if 'token' not in kwargs:
            kwargs['token'] = uuid.uuid4().hex
        model = super().create(**kwargs)
        return model

    @classmethod
    @safe_db_query
    def create_or_update(self, trigger_config: Trigger):
        try:
            existing_trigger = PipelineSchedule.query.filter(
                self.name == trigger_config.name,
                self.pipeline_uuid == trigger_config.pipeline_uuid,
            ).one_or_none()
        except Exception:
            traceback.print_exc()
            existing_trigger = None

        kwargs = dict(
            name=trigger_config.name,
            pipeline_uuid=trigger_config.pipeline_uuid,
            schedule_type=trigger_config.schedule_type,
            start_time=trigger_config.start_time,
            schedule_interval=trigger_config.schedule_interval,
            status=trigger_config.status,
            variables=trigger_config.variables,
            sla=trigger_config.sla,
            settings=trigger_config.settings,
        )

        if existing_trigger:
            existing_trigger.update(**kwargs)
        else:
            self.create(**kwargs)

    def current_execution_date(self) -> datetime:
        if self.schedule_interval is None:
            return None

        now = datetime.now(timezone.utc)
        if self.schedule_interval == '@once':
            return now
        elif self.schedule_interval == '@daily':
            return now.replace(second=0, microsecond=0, minute=0, hour=0)
        elif self.schedule_interval == '@hourly':
            return now.replace(second=0, microsecond=0, minute=0)
        elif self.schedule_interval == '@weekly':
            return now.replace(second=0, microsecond=0, minute=0, hour=0) - \
                timedelta(days=now.weekday())
        elif self.schedule_interval == '@monthly':
            return now.replace(second=0, microsecond=0, minute=0, hour=0, day=1)
        else:
            cron_itr = croniter(self.schedule_interval, now)
            return cron_itr.get_prev(datetime)

    @safe_db_query
    def should_schedule(self) -> bool:
        if self.status != ScheduleStatus.ACTIVE:
            return False

        if self.start_time is not None and compare(datetime.now(), self.start_time) == -1:
            return False

        try:
            Pipeline.get(self.pipeline_uuid)
        except Exception:
            return False

        if self.schedule_interval == '@once':
            pipeline_run_count = len(self.pipeline_runs)
            if pipeline_run_count == 0:
                return True
            executor_count = self.pipeline.executor_count
            # Used by streaming pipeline to launch multiple executors
            if executor_count > 1 and pipeline_run_count < executor_count:
                return True
        else:
            """
            TODO: Implement other schedule interval checks
            """
            current_execution_date = self.current_execution_date()
            if current_execution_date is None:
                return False
            if not find(
                lambda x: compare(
                    x.execution_date.replace(tzinfo=pytz.UTC),
                    current_execution_date,
                ) == 0,
                self.pipeline_runs
            ):
                return True
        return False


class PipelineRun(BaseModel):
    class PipelineRunStatus(str, enum.Enum):
        INITIAL = 'initial'
        RUNNING = 'running'
        COMPLETED = 'completed'
        FAILED = 'failed'
        CANCELLED = 'cancelled'

    pipeline_schedule_id = Column(Integer, ForeignKey('pipeline_schedule.id'))
    pipeline_uuid = Column(String(255), index=True)
    execution_date = Column(DateTime(timezone=True), index=True)
    status = Column(Enum(PipelineRunStatus), default=PipelineRunStatus.INITIAL, index=True)
    completed_at = Column(DateTime(timezone=True))
    variables = Column(JSON)
    passed_sla = Column(Boolean, default=False)
    event_variables = Column(JSON)
    metrics = Column(JSON)
    backfill_id = Column(Integer, ForeignKey('backfill.id'))

    pipeline_schedule = relationship(PipelineSchedule, back_populates='pipeline_runs')
    block_runs = relationship('BlockRun', back_populates='pipeline_run')
    backfill = relationship('Backfill', back_populates='pipeline_runs')

    def __repr__(self):
        return f'PipelineRun(id={self.id}, pipeline_uuid={self.pipeline_uuid},'\
               f' execution_date={self.execution_date})'

    @property
    def block_runs_count(self) -> int:
        return len(self.block_runs)

    @property
    def execution_partition(self) -> str:
        if self.execution_date is None:
            return str(self.pipeline_schedule_id)
        else:
            return '/'.join([
                        str(self.pipeline_schedule_id),
                        self.execution_date.strftime(format='%Y%m%dT%H%M%S'),
                    ])

    @property
    def pipeline(self) -> 'Pipeline':
        return Pipeline.get(self.pipeline_uuid)

    @property
    def logs(self):
        return LoggerManagerFactory.get_logger_manager(
            pipeline_uuid=self.pipeline_uuid,
            partition=self.execution_partition,
            repo_config=self.pipeline.repo_config,
        ).get_logs()

    async def logs_async(self):
        return await LoggerManagerFactory.get_logger_manager(
            pipeline_uuid=self.pipeline_uuid,
            partition=self.execution_partition,
            repo_config=self.pipeline.repo_config,
        ).get_logs_async()

    @property
    def pipeline_schedule_name(self):
        return self.pipeline_schedule.name

    @property
    def pipeline_schedule_token(self):
        return self.pipeline_schedule.token

    @property
    def pipeline_schedule_type(self):
        return self.pipeline_schedule.schedule_type

    @classmethod
    @safe_db_query
    def active_runs(
        self,
        pipeline_uuids: List[str] = None,
        include_block_runs: bool = False,
    ) -> List['PipelineRun']:
        query = self.query.filter(self.status == self.PipelineRunStatus.RUNNING)
        if pipeline_uuids is not None:
            query = query.filter(PipelineRun.pipeline_uuid.in_(pipeline_uuids))
        if include_block_runs:
            query = query.options(joinedload(PipelineRun.block_runs))
        return query.all()

    @classmethod
    def create(self, create_block_runs: bool = True, **kwargs) -> 'PipelineRun':
        pipeline_run = super().create(**kwargs)
        pipeline_uuid = kwargs.get('pipeline_uuid')
        if pipeline_uuid is not None and create_block_runs:
            pipeline_run.create_block_runs()

        return pipeline_run

    @classmethod
    @safe_db_query
    def in_progress_runs(
        self,
        pipeline_schedules: List[int],
    ):
        return self.query.filter(
            PipelineRun.pipeline_schedule_id.in_(pipeline_schedules),
            PipelineRun.status.in_([
                self.PipelineRunStatus.INITIAL,
                self.PipelineRunStatus.RUNNING,
            ]),
            PipelineRun.passed_sla.is_(False),
        ).all()

    @safe_db_query
    def create_block_run(
        self,
        block_uuid: str,
        skip_if_exists: bool = False,
        **kwargs,
    ) -> 'BlockRun':
        if skip_if_exists:
            br = BlockRun.get(
                pipeline_run_id=self.id,
                block_uuid=block_uuid,
            )
            if br is not None:
                return br
        return BlockRun.create(
            block_uuid=block_uuid,
            pipeline_run_id=self.id,
            **kwargs,
        )

    def create_block_runs(self) -> List['BlockRun']:
        blocks = self.pipeline.get_executable_blocks()

        arr = []
        for block in blocks:
            ancestors = get_all_ancestors(block)
            if len(block.upstream_blocks) == 0 or not find(is_dynamic_block, ancestors):
                arr.append(block)

        return [self.create_block_run(b.uuid) for b in arr]

    def any_blocks_failed(self) -> bool:
        return any(
            b.status == BlockRun.BlockRunStatus.FAILED
            for b in self.block_runs
        )

    def all_blocks_completed(self, include_failed_blocks: bool = False) -> bool:
        statuses = [BlockRun.BlockRunStatus.COMPLETED]
        if include_failed_blocks:
            statuses.extend([
                BlockRun.BlockRunStatus.FAILED,
                BlockRun.BlockRunStatus.UPSTREAM_FAILED,
            ])
        return all(b.status in statuses for b in self.block_runs)


class BlockRun(BaseModel):
    class BlockRunStatus(str, enum.Enum):
        INITIAL = 'initial'
        QUEUED = 'queued'
        RUNNING = 'running'
        COMPLETED = 'completed'
        FAILED = 'failed'
        CANCELLED = 'cancelled'
        UPSTREAM_FAILED = 'upstream_failed'

    pipeline_run_id = Column(Integer, ForeignKey('pipeline_run.id'))
    block_uuid = Column(String(255))
    status = Column(Enum(BlockRunStatus), default=BlockRunStatus.INITIAL)
    started_at = Column(DateTime(timezone=True))
    completed_at = Column(DateTime(timezone=True))
    metrics = Column(JSON)

    pipeline_run = relationship(PipelineRun, back_populates='block_runs')

    @property
    def logs(self):
        pipeline = Pipeline.get(self.pipeline_run.pipeline_uuid)
        return LoggerManagerFactory.get_logger_manager(
            pipeline_uuid=pipeline.uuid,
            block_uuid=clean_name(self.block_uuid),
            partition=self.pipeline_run.execution_partition,
            repo_config=pipeline.repo_config,
        ).get_logs()

    async def logs_async(self):
        pipeline = await Pipeline.get_async(self.pipeline_run.pipeline_uuid)
        return await LoggerManagerFactory.get_logger_manager(
            pipeline_uuid=pipeline.uuid,
            block_uuid=clean_name(self.block_uuid),
            partition=self.pipeline_run.execution_partition,
            repo_config=pipeline.repo_config,
        ).get_logs_async()

    @classmethod
    @safe_db_query
    def batch_update_status(self, block_run_ids: List[int], status):
        BlockRun.query.filter(BlockRun.id.in_(block_run_ids)).update({
            BlockRun.status: status
        }, synchronize_session=False)
        db_connection.session.commit()

    @classmethod
    @safe_db_query
    def get(self, pipeline_run_id: int = None, block_uuid: str = None) -> 'BlockRun':
        block_runs = self.query.filter(
            BlockRun.pipeline_run_id == pipeline_run_id,
            BlockRun.block_uuid == block_uuid,
        ).all()
        if len(block_runs) > 0:
            return block_runs[0]
        return None

    def get_outputs(self, sample_count: int = None) -> List[Dict]:
        pipeline = Pipeline.get(self.pipeline_run.pipeline_uuid)
        block = pipeline.get_block(self.block_uuid)
        return block.get_outputs(
            execution_partition=self.pipeline_run.execution_partition,
            sample_count=sample_count,
            block_uuid=self.block_uuid,
        )


class EventMatcher(BaseModel):
    class EventType(str, enum.Enum):
        AWS_EVENT = 'aws_event'

    event_type = Column(Enum(EventType), default=EventType.AWS_EVENT)
    name = Column(String(255))
    pattern = Column(JSON)

    pipeline_schedules = relationship(
        'PipelineSchedule',
        secondary=pipeline_schedule_event_matcher_association_table,
        back_populates='event_matchers',
    )

    def __repr__(self):
        return f'EventMatcher(id={self.id}, name={self.name}, pattern={self.pattern})'

    @classmethod
    def active_event_matchers(self) -> List['EventMatcher']:
        return self.query.filter(
            EventMatcher.pipeline_schedules.any(
                PipelineSchedule.status == ScheduleStatus.ACTIVE
            )
        ).all()

    @classmethod
    def upsert_batch(self, event_matchers_payload):
        keys_to_ignore = [
            'created_at',
            'id',
            'updated_at',
        ]

        new_arr = []
        existing_arr = []

        pipeline_schedule_ids = []
        for payload in event_matchers_payload:
            pipeline_schedule_ids += payload.get('pipeline_schedule_ids', [])
            if payload.get('id'):
                existing_arr.append(payload)
            else:
                new_arr.append(payload)

        pipeline_schedules_by_id = index_by(
            lambda x: x.id,
            PipelineSchedule.query.filter(
                PipelineSchedule.id.in_(pipeline_schedule_ids),
            ).all(),
        )

        event_matchers_and_pipeline_schedule_ids = []
        event_matchers_by_id = index_by(
            lambda x: x.id,
            self.query.filter(
                self.id.in_([p['id'] for p in existing_arr]),
            ).all(),
        )
        for payload in existing_arr:
            ids = payload.pop('pipeline_schedule_ids', None)
            event_matcher = event_matchers_by_id[payload['id']]
            event_matcher.update(**ignore_keys(payload, keys_to_ignore))
            event_matchers_and_pipeline_schedule_ids.append((event_matcher, ids))

        for payload in new_arr:
            ids = payload.pop('pipeline_schedule_ids', None)
            event_matcher = self.create(**ignore_keys(payload, keys_to_ignore))
            event_matchers_and_pipeline_schedule_ids.append((event_matcher, ids))

        for event_matcher, ids in event_matchers_and_pipeline_schedule_ids:
            if ids is not None:
                ps = [pipeline_schedules_by_id[i] for i in [int(i) for i in ids]]
                event_matcher.update(pipeline_schedules=ps)

            if event_matcher.event_type == EventMatcher.EventType.AWS_EVENT:
                from mage_ai.services.aws.events.events import update_event_rule_targets
                # For AWS event, update related AWS infra (add trigger to lambda function)
                update_event_rule_targets(event_matcher.name)

        return [t[0] for t in event_matchers_and_pipeline_schedule_ids]

    def active_pipeline_schedules(self) -> List[PipelineSchedule]:
        return [p for p in self.pipeline_schedules
                if p.status == ScheduleStatus.ACTIVE]

    def match(self, config: Dict) -> bool:
        def __match_dict(sub_pattern, sub_config):
            if type(sub_pattern) is not dict or type(sub_config) is not dict:
                return False
            for k in sub_pattern.keys():
                if k not in sub_config:
                    return False
                v = sub_pattern[k]
                if type(v) is list:
                    if sub_config[k] not in v:
                        return False
                elif not __match_dict(v, sub_config[k]):
                    return False
            return True
        return __match_dict(self.pattern, config)


class Backfill(BaseModel):
    class IntervalType(str, enum.Enum):
        SECOND = 'second'
        MINUTE = 'minute'
        HOUR = 'hour'
        DAY = 'day'
        WEEK = 'week'
        MONTH = 'month'
        YEAR = 'year'
        CUSTOM = 'custom'

    class Status(str, enum.Enum):
        INITIAL = 'initial'
        RUNNING = 'running'
        COMPLETED = 'completed'
        FAILED = 'failed'
        CANCELLED = 'cancelled'

    block_uuid = Column(String(255), default=None)
    completed_at = Column(DateTime(timezone=True), default=None)
    created_at = Column(DateTime(timezone=True), server_default=func.now())
    end_datetime = Column(DateTime(timezone=True), default=None)
    failed_at = Column(DateTime(timezone=True), default=None)
    interval_type = Column(Enum(IntervalType), default=None)
    interval_units = Column(Integer, default=None)
    metrics = Column(JSON)
    name = Column(String(255))
    pipeline_schedule = relationship(PipelineSchedule, back_populates='backfills')
    pipeline_schedule_id = Column(Integer, ForeignKey('pipeline_schedule.id'))
    pipeline_uuid = Column(String(255))
    start_datetime = Column(DateTime(timezone=True), default=None)
    started_at = Column(DateTime(timezone=True), default=None)
    status = Column(Enum(Status), default=None)
    updated_at = Column(DateTime(timezone=True), server_default=func.now(), onupdate=func.now())
    variables = Column(JSON, default=None)

    pipeline_runs = relationship('PipelineRun', back_populates='backfill')

    @classmethod
    @safe_db_query
    def filter(self, pipeline_schedule_ids: List = None):
        if pipeline_schedule_ids is not None:
            return Backfill.query.filter(
                Backfill.pipeline_schedule_id.in_(pipeline_schedule_ids),
            )
        return []



from dataclasses import dataclass
from mage_ai.shared.config import BaseConfig
from enum import Enum


class QueueType(str, Enum):
    CELERY = 'celery'
    PROCESS = 'process'


@dataclass
class QueueConfig(BaseConfig):
    queue_type: QueueType = QueueType.PROCESS
    concurrency: int = 20



from botocore.config import Config
from functools import reduce
from mage_ai.services.aws.ecs.config import EcsConfig
from mage_ai.services.aws.ecs.ecs import list_tasks, run_task, stop_task
from mage_ai.shared.array import find
from mage_ai.shared.hash import dig
from typing import Dict, List

import boto3
import json
import os


class EcsTaskManager:
    def __init__(self, cluster_name):
        self.cluster_name = cluster_name

        self.metadata_file = os.path.join(
            os.getcwd(),
            'instance_metadata.json',
        )

        if not os.path.exists(self.metadata_file):
            self.instance_metadata = {}

    @property
    def instance_metadata(self):
        metadata = {}
        with open(self.metadata_file, 'r', encoding='utf-8') as file:
            metadata = json.load(file)
        return metadata

    @instance_metadata.setter
    def instance_metadata(self, metadata):
        with open(self.metadata_file, 'w', encoding='utf-8') as file:
            json.dump(metadata, file)

    def list_tasks(self):
        region_name = os.getenv('AWS_REGION_NAME', 'us-west-2')
        config = Config(region_name=region_name)
        ec2_client = boto3.client('ec2', config=config)

        response = list_tasks(self.cluster_name)['tasks']
        network_interfaces = self.__get_network_interfaces(response, ec2_client)

        tasks = []
        for task in response:
            public_ip = dig(network_interfaces.get(task['taskArn']), 'Association.PublicIp')

            tags = task['tags']
            name = find(lambda tag: tag.get('key') == 'name', tags)

            tasks.append(dict(
                ip=public_ip,
                name=name.get('value') if name is not None else None,
                status=task['lastStatus'],
                task_arn=task['taskArn'],
                type=task['launchType'],
            ))

        running_instance_names = set(map(lambda x: x['name'], tasks))

        stopped_instance_names = [
            name for name in list(self.instance_metadata.keys())
            if name not in running_instance_names
        ]
        stopped_instances = \
            list(
                map(
                    lambda name: {'name': name, 'status': 'STOPPED'},
                    stopped_instance_names
                )
            )

        return tasks + stopped_instances

    def create_task(self, name: str, task_definition: str, container_name: str):
        region_name = os.getenv('AWS_REGION_NAME', 'us-west-2')
        config = Config(region_name=region_name)
        ec2_client = boto3.client('ec2', config=config)

        # create new task
        task = find(
            lambda task: task.get('lastStatus') == 'RUNNING',
            list_tasks(self.cluster_name)['tasks'],
        )
        network_interface = self.__get_network_interfaces([task], ec2_client)[task['taskArn']]

        subnets = [network_interface['SubnetId']]
        security_groups = [g['GroupId'] for g in network_interface['Groups']]

        ecs_config = EcsConfig(
            task_definition,
            container_name,
            self.cluster_name,
            security_groups=security_groups,
            subnets=subnets,
            tags=[
                {
                    'key': 'name',
                    'value': name,
                }
            ],
        )

        self.instance_metadata = {
            **self.instance_metadata,
            name: dict()
        }

        return run_task(f'mage start {name}', ecs_config=ecs_config)

    def stop_task(self, task_arn: str):
        return stop_task(task_arn, self.cluster_name)

    def delete_task(self, name, task_arn: str = None):
        if task_arn:
            self.stop_task(task_arn)

        updated_metadata = self.instance_metadata

        if name in updated_metadata:
            del updated_metadata[name]
            self.instance_metadata = updated_metadata

    def __get_network_interface_id(self, task):
        if task.get('lastStatus') != 'RUNNING':
            return None

        attachment = \
            find(lambda a: a['type'] == 'ElasticNetworkInterface', task.get('attachments', []))
        network_interface = \
            find(lambda d: d['name'] == 'networkInterfaceId', attachment.get('details', []))
        return network_interface.get('value', None)

    def __get_network_interfaces(self, tasks: List, ec2_client) -> Dict:
        task_mapping = dict()
        for task in tasks:
            nii = self.__get_network_interface_id(task)
            if nii is not None:
                task_mapping[task['taskArn']] = nii

        network_interface_ids = list(task_mapping.values())

        network_interfaces = ec2_client.describe_network_interfaces(
            NetworkInterfaceIds=network_interface_ids
        )['NetworkInterfaces']

        def aggregate(obj, task):
            task_arn = task['taskArn']
            if task_arn in task_mapping:
                obj[task_arn] = find(
                    lambda i: i['NetworkInterfaceId'] == task_mapping[task_arn],
                    network_interfaces,
                )
            return obj

        return reduce(aggregate, tasks, {})



from datetime import datetime, timedelta
from mage_ai.services.stitch.config import StitchConfig
from mage_ai.services.stitch.constants import (
    DEFAULT_POLL_INTERVAL,
    DEFAULT_POLL_TIMEOUT,
    STITCH_BASE_URL,
)
from mage_ai.shared.http_client import HttpClient
from typing import Dict, List, Optional, Union
import requests
import time

LOG_TEXT_FOR_STREAMS_WITH_EXTRACTED_ROWS = \
    'com.stitchdata.target-stitch-avro.flush-pipeline - send-stream-record-count-impl'


class StitchClient(HttpClient):
    """
    API doc: https://www.stitchdata.com/docs/developers/stitch-connect/api
    """

    BASE_URL = STITCH_BASE_URL

    def __init__(self, config: Union[Dict, StitchConfig]):
        if type(config) is dict:
            self.config = StitchConfig.load(config=config)
        else:
            self.config = config
        self.headers = {
            'Content-Type': 'application/json',
            'Authorization': f'Bearer {self.config.access_token}',
        }

    def list_sources(self):
        return self.make_request('/sources')

    def get_source(self, source_id: int):
        return self.make_request(f'/sources/{source_id}')

    def list_streams(self, source_id: int, selected: bool = True):
        streams = self.make_request(f'/sources/{source_id}/streams')
        if selected:
            streams = [s for s in streams if s['selected']]
        return streams

    def list_destinations(self):
        return self.make_request('/destinations')

    def get_destination(self, destination_id: int):
        return self.make_request(f'/destinations/{destination_id}')

    def list_extractions(self, stitch_client_id: int, page: int = 1):
        return self.make_request(f'/{stitch_client_id}/extractions', params=dict(page=page))

    def list_loads(self, stitch_client_id: int, page: int = 1):
        return self.make_request(f'/{stitch_client_id}/loads', params=dict(page=page))

    def check_sync_completion(
        self,
        source_id: str,
        job_name: str,
        poll_interval: float = DEFAULT_POLL_INTERVAL,
        poll_timeout: Optional[float] = DEFAULT_POLL_TIMEOUT,
        autocomplete_after_seconds: int = None,
    ):
        source = self.get_source(source_id)
        stitch_client_id = source['stitch_client_id']

        # Poll status for Extraction job
        poll_start = datetime.now()
        extraction_completion_time = None
        while True:
            extractions = self.list_extractions(stitch_client_id)['data']
            extractions = [e for e in extractions if e['job_name'] == job_name]
            if len(extractions) == 0:
                print(
                    f'Polling Stitch extraction status for source {source_id}. '
                    'Current status: running.'
                )
            else:
                extraction = extractions[0]

                if extraction['discovery_exit_status'] is None:
                    print(
                        f'Polling Stitch extraction status for source {source_id}. '
                        'Current status: running.'
                    )
                elif extraction['discovery_exit_status'] == 0:
                    extraction_completion_time = extraction['completion_time']
                    print(f'Extraction for source {source_id} completed.')
                    break
                else:
                    error_message = extraction['discovery_description']
                    raise Exception(
                        f'Extraction for source {source_id} failed with '
                        f'message: \"{error_message}\".'
                    )
            if (
                poll_timeout
                and datetime.now()
                > poll_start + timedelta(seconds=poll_timeout)
            ):
                raise Exception(
                    f'Extraction for source {source_id} times out after '
                    f'{datetime.now() - poll_start}.'
                )
            time.sleep(poll_interval)

        # Poll status for Load job
        poll_start = datetime.now()
        stream_names = self.__get_streams_extracted_from_extraction_job(stitch_client_id, job_name)
        while True:
            succeeded_streams = set()
            total_loads_count = None
            current_count = 0
            page = 1
            while True:
                # Fetch loads with pagination
                loads_response = self.list_loads(stitch_client_id, page=page)
                loads = loads_response['data']
                if total_loads_count is None:
                    total_loads_count = loads_response['total']
                current_count += len(loads)
                loads = [load for load in loads if load['source_name'] == source['name']
                         and load['stream_name'] in stream_names]
                for load in loads:
                    if load['error_state'] is not None:
                        error_message = \
                            load['error_state']['notification_data']['warehouse_message']
                        raise Exception(
                            f"Failed to load data for stream {load['stream_name']} with "
                            f"message: \"{error_message}\"."
                        )
                    elif load['last_batch_loaded_at'] >= extraction_completion_time:
                        succeeded_streams.add(load['stream_name'])
                page += 1
                if current_count >= total_loads_count:
                    break

            succeeded_streams = list(succeeded_streams)
            total_streams = len(stream_names)
            completed_streams = len(succeeded_streams)

            if completed_streams == total_streams:
                print(f'Finish loading data for all streams: {succeeded_streams}.')
                break
            elif autocomplete_after_seconds and \
                    datetime.now().timestamp() - autocomplete_after_seconds >= \
                    poll_start.timestamp():
                print(f'Automatically setting job as complete after {autocomplete_after_seconds} '
                      'seconds.')
                break
            else:
                percent_complete = round(
                    100 * (completed_streams / total_streams),
                    2,
                ) if total_streams else 0
                running_streams = [s for s in stream_names if s not in succeeded_streams]
                print(
                    f'Polling Stitch load status for source {source_id}: {percent_complete}% '
                    f'({completed_streams}/{total_streams}). '
                    f'Completed streams: {succeeded_streams}. '
                    f'Running streams: {running_streams}.'
                )
            if (
                poll_timeout
                and datetime.now()
                > poll_start + timedelta(seconds=poll_timeout)
            ):
                raise Exception(
                    f'Load for source {source_id} times out after {datetime.now() - poll_start}.'
                )
            time.sleep(poll_interval)

    def start_replication_job(
        self,
        source_id: int,
        poll_interval: float = DEFAULT_POLL_INTERVAL,
        poll_timeout: Optional[float] = DEFAULT_POLL_TIMEOUT,
        autocomplete_after_seconds: int = None,
        disable_polling: bool = False,
    ):
        response = self.make_request(f'/sources/{source_id}/sync', method='POST', payload=dict())
        if 'error' in response:
            raise Exception(response['error']['message'])
        job_name = response['job_name']
        print(f'Start replication job for source {source_id}. Job name: {job_name}.')

        if disable_polling:
            return response

        self.check_sync_completion(
            source_id,
            job_name,
            poll_interval=poll_interval,
            poll_timeout=poll_timeout,
            autocomplete_after_seconds=autocomplete_after_seconds,
        )

        return response

    def __get_logs_for_extraction(self, stitch_client_id: int, job_name: str) -> str:
        url = f'{STITCH_BASE_URL}/{stitch_client_id}/extractions/{job_name}'
        response = requests.get(url, headers=self.headers)
        return response.text

    def __get_streams_extracted_from_extraction_job(
        self,
        stitch_client_id: int,
        job_name: str,
    ) -> List[str]:
        logs = self.__get_logs_for_extraction(stitch_client_id, job_name)
        streams = set()
        for line in logs.split('\n'):
            if LOG_TEXT_FOR_STREAMS_WITH_EXTRACTED_ROWS in line:
                stream_name = line.split(LOG_TEXT_FOR_STREAMS_WITH_EXTRACTED_ROWS)[1].strip()
                streams.add(stream_name)
        return streams
